from __future__ import annotations

import sys
import os

import _setup_test_env  # noqa
import returnn.sprint.interface as SprintAPI
from returnn.tf.engine import Engine
from tempfile import mkdtemp
from returnn.config import Config
import shutil
import numpy


def install_sigint_handler():
    import signal

    def signal_handler(signal, frame):
        print("\nSIGINT at:")
        better_exchook.print_tb(tb=frame, file=sys.stdout)
        print("")

        # It's likely that SIGINT was caused by Util.interrupt_main().
        # We might have a stacktrace from there.
        if getattr(sys, "exited_frame", None) is not None:
            print("interrupt_main via:")
            better_exchook.print_tb(tb=sys.exited_frame, file=sys.stdout)
            print("")
            sys.exited_frame = None
            # Normal exception instead so that Nose will catch it.
            raise Exception("Got SIGINT!")
        else:
            print("\nno sys.exited_frame\n")
            # Normal SIGINT. Normal Nose exit.
            if old_action:
                old_action()
            else:
                raise KeyboardInterrupt

    old_action = signal.signal(signal.SIGINT, signal_handler)


install_sigint_handler()


def test_forward():
    tmpdir = mkdtemp("returnn-test-sprint")
    olddir = os.getcwd()
    os.chdir(tmpdir)

    from returnn.datasets.generating import DummyDataset

    seq_len = 5
    n_data_dim = 2
    n_classes_dim = 3
    train_data = DummyDataset(input_dim=n_data_dim, output_dim=n_classes_dim, num_seqs=4, seq_len=seq_len)
    train_data.init_seq_order(epoch=1)
    cv_data = DummyDataset(input_dim=n_data_dim, output_dim=n_classes_dim, num_seqs=2, seq_len=seq_len)
    cv_data.init_seq_order(epoch=1)

    config = "\n".join(
        [
            "#!rnn.py",
            "use_tensorflow = True",
            "model = '%s/model'" % tmpdir,
            "num_outputs = %i" % n_classes_dim,
            "num_inputs = %i" % n_data_dim,
            'network = {"output": {"class": "softmax", "loss": "ce", "from": "data:data"}}',
            "num_epochs = 2",
        ]
    )

    open("config_write", "w").write(config)
    open("config", "w").write(config + "\nload= '%s/model'" % tmpdir)

    config = Config()
    config.load_file("config_write")
    engine = Engine(config=config)
    engine.init_train_from_config(config=config, train_data=train_data, dev_data=cv_data, eval_data=None)
    engine.epoch = 1
    engine.save_model(engine.get_epoch_model_filename())
    Engine._epoch_model = None

    # Reset engine
    from returnn.util.basic import BackendEngine

    BackendEngine.selected_engine = None

    inputDim = 2
    outputDim = 3
    SprintAPI.init(
        inputDim=inputDim,
        outputDim=outputDim,
        config="action:forward,configfile:config,epoch:1",
        targetMode="forward-only",
    )
    assert isinstance(SprintAPI.engine, Engine)
    print("used data keys via net:", SprintAPI.engine.network.get_used_data_keys())

    features = numpy.array([[0.1, 0.2], [0.2, 0.3], [0.3, 0.4], [0.4, 0.5]])
    seq_len = features.shape[0]
    posteriors = SprintAPI._forward("segment1", features.T).T
    assert posteriors.shape == (seq_len, outputDim)

    SprintAPI.exit()

    os.chdir(olddir)
    shutil.rmtree(tmpdir)



"""
Collection of generic utilities related to signal processing
"""

import numpy


def greenwood_function(x, scaling_constant=165.4, constant_of_integration=0.88, slope=2.1):
    """
    Greenwood function, convert fractional length to frequency, see
    Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
    ICASSP 2007 and also
    https://en.wikipedia.org/wiki/Greenwood_function
    The default values are taken from RASR, see https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc
    They correspond to the recommended values for human data according to
    Greenwood, Donald D. "A cochlear frequency‐position function for several species—29 years later."
    The Journal of the Acoustical Society of America, 1990

    :param float x: fractional length
    :param float scaling_constant: A in [1]
    :param float constant_of_integration: k in [1]
    :param float slope: a in [1]
    :return: frequency corresponding to given fractional length
    :rtype: float
    """
    assert 0 <= x <= 1, "fractional length has to be between 0 and 1"
    return scaling_constant * (numpy.power(10.0, slope * x) - constant_of_integration)


def inv_greenwood_function(freq, scaling_constant=165.4, constant_of_integration=0.88, slope=2.1):
    """
    Inverse greenwood function, convert frequency to fractional length, see
    Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
    ICASSP 2007 and also
    https://en.wikipedia.org/wiki/Greenwood_function
    The default values are taken from RASR, see https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc
    They correspond to the recommended values for human data according to
    Greenwood, Donald D. "A cochlear frequency‐position function for several species—29 years later."
    The Journal of the Acoustical Society of America, 1990

    :param float freq: frequency
    :param float scaling_constant: A in [1]
    :param float constant_of_integration: k in [1]
    :param float slope: a in [1]
    :return: fractional length corresponding to given frequency
    :rtype: float
    """
    return numpy.log10(freq / scaling_constant + constant_of_integration) / slope


class GammatoneFilterbank:
    """
    Class representing a gammatone filterbank.
    Based on
    [1] Schlueter, Ralf, et al. "Gammatone features and feature combination for large vocabulary speech recognition."
    ICASSP 2007
    """

    def __init__(self, num_channels, length, sample_rate=16000, freq_max=7500.0, freq_min=100.0, normalization=True):
        """
        :param int num_channels: number of filters
        :param int|float length: length of FIR filters in seconds
        :param int sample_rate: sample rate of audio signal in Hz
        :param float freq_max: maximum frequency of filterbank
        :param float freq_min: minimum frequency of filterbank
        :param bool normalization: normalize filterbanks to maximum frequency response of 0 dB
        """
        self.num_channels = num_channels
        self.length = length
        self.sample_rate = sample_rate
        self.freq_max = freq_max
        self.freq_min = freq_min
        self.normalization = normalization

    def get_gammatone_filterbank(self):
        """
        Returns an array with the parameters of the gammatone filterbank

        :return: gammatone filterbank of shape (self.length * self.sample_rate, self.num_channels)
        :rtype: numpy.array
        """
        center_freqs = self.center_frequencies(self.num_channels, self.freq_max, self.freq_min)
        fbank = []
        for freq in center_freqs:
            fbank.append(self.gammatone_impulse_response(freq, self.length, self.sample_rate))
        fbank = numpy.vstack(fbank)
        if self.normalization:
            fbank = self.normalize_filters(fbank)
        fbank = numpy.transpose(fbank)
        return fbank

    @staticmethod
    def center_frequencies(num_channels, freq_max, freq_min):
        """
        Determine center frequencies for gammatone filterbank

        :param int num_channels: number of filters
        :param float freq_max: maximum frequency of filterbank
        :param float freq_min: minimum frequency of filterbank
        :return: center frequencies
        :rtype: numpy.array
        """
        x_min = inv_greenwood_function(freq_min)
        x_max = inv_greenwood_function(freq_max)
        x_linear = numpy.linspace(x_min, x_max, num_channels)
        freq_center = numpy.vectorize(greenwood_function)(x_linear)
        return freq_center

    @staticmethod
    def bandwidth_by_center_frequency(freq, lin_approx_coeff=24.7, quality_factor=9.264491981582191):
        """
        Get bandwidth (named B in [1]) by center frequency using a linear approximation of the equivalent rectangular
        bandwidth (ERB) from
        Glasberg, Brian R., and Brian CJ Moore. "Derivation of auditory filter shapes from notched-noise data."
        Hearing research, 1990
        The default values are taken from there and are also used in RASR, see
        https://github.com/rwth-i6/rasr/blob/master/src/Signal/GammaTone.cc

        :param float freq: center frequency
        :param float lin_approx_coeff: coefficient for the linear approximation of the ERB
        :param float quality_factor: audiological (ERB) based filter quality factor
        :return: bandwidth
        :rtype: float
        """
        return lin_approx_coeff * (1 / (lin_approx_coeff * quality_factor) * freq + 1.0)

    def gammatone_impulse_response(
        self, f_center, length, sample_rate, output_gain=1.0, filter_order=4, phase_shift=0.0
    ):
        """
        Compute gammatone impulse response based on [1]

        :param float f_center: center frequency
        :param int|float length: length of finite impulse response in seconds
        :param int sample_rate: sample rate of audio signal in Hz
        :param float output_gain: output gain, named k in [1]
        :param int filter_order: order of filter, named n in [1]
        :param float phase_shift: phase shift, named phi in [1]
        :return: gammatone impulse response
        :rtype: numpy.array
        """
        bandwidth = self.bandwidth_by_center_frequency(f_center)  # bandwidth (duration of impulse response)
        num_samples = int(numpy.floor(sample_rate * length))
        t = numpy.linspace(1.0 / sample_rate, length, num_samples)
        return (
            output_gain
            * numpy.power(t, filter_order - 1)
            * numpy.exp(-2 * numpy.pi * bandwidth * t)
            * numpy.cos(2 * numpy.pi * f_center * t + phase_shift)
        )

    @staticmethod
    def normalize_filters(filters):
        """
        Normalize filterbank such that the maximum frequency response is 0 dB

        :param numpy.array filters: filterbank with shape number_channels x filter_length
        :return: normalized filterbank
        :rtype: numpy.array
        """
        from scipy import signal

        for filt in range(filters.shape[0]):
            _, f_resp = signal.freqz(filters[filt, :])
            filters[filt, :] = filters[filt, :] / numpy.max(numpy.abs(f_resp))
        return filters



"""
Loss functions
"""

from __future__ import annotations
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["cross_entropy", "ctc_loss", "edit_distance"]


def cross_entropy(
    *,
    estimated: Tensor,
    target: Tensor,
    axis: Dim,
    estimated_type: str,
) -> Tensor:
    """
    ``target`` is supposed to be in probability space (normalized). It can also be sparse, i.e. contain class indices.
    ``estimated`` can be probs, log-probs or logits, specified via ``estimated_type``.

    Assuming both are in probability space, the cross entropy is:

        H(target,estimated) = -reduce_sum(target * log(estimated), axis=axis)
                            = -matmul(target, log(estimated), reduce=axis)

    In case you want label smoothing, you can use e.g.::

        ce = nn.cross_entropy(
            target=nn.label_smoothing(target, 0.1),
            estimated=estimated)

    :param estimated: probs, log-probs or logits, specified via ``estimated_type``
    :param target: probs, normalized, can also be sparse
    :param axis: class labels dim over which softmax is computed
    :param estimated_type: "probs", "log-probs" or "logits"
    :return: cross entropy (same Dims as 'estimated' but without 'axis')
    """

    if estimated_type == "logits":
        # This is a common case and most backends provide optimized functions for it.
        # noinspection PyProtectedMember
        return estimated._raw_backend.softmax_cross_entropy_with_logits(logits=estimated, targets=target, axis=axis)
    if estimated_type == "probs":
        log_prob = rf.log(estimated)  # TODO: make numerically stable
    elif estimated_type == "log-probs":
        log_prob = estimated
    else:
        raise ValueError("estimated_type must be 'probs', 'log-probs' or 'logits'")
    if target.sparse_dim:
        return -rf.gather(log_prob, indices=target, axis=axis)
    return -rf.matmul(target, log_prob, reduce=axis)


def ctc_loss(
    *,
    logits: Tensor,
    logits_normalized: bool = False,
    targets: Tensor,
    input_spatial_dim: Dim,
    targets_spatial_dim: Dim,
    blank_index: int,
    max_approx: bool = False,
) -> Tensor:
    """
    Calculates the CTC loss.

    Internally, this uses :func:`returnn.tf.native_op.ctc_loss`
    which is equivalent to tf.nn.ctc_loss but more efficient.

    Output is of shape [B].

    :param logits: (before softmax). shape [B...,input_spatial,C]
    :param logits_normalized: whether the logits are already normalized (e.g. via log-softmax)
    :param targets: sparse. shape [B...,targets_spatial] -> C
    :param input_spatial_dim: spatial dim of input logits
    :param targets_spatial_dim: spatial dim of targets
    :param blank_index: vocab index of the blank symbol
    :param max_approx: if True, use max instead of sum over alignments (max approx, Viterbi)
    :return: loss shape [B...]
    """
    # noinspection PyProtectedMember
    return logits._raw_backend.ctc_loss(
        logits=logits,
        logits_normalized=logits_normalized,
        targets=targets,
        input_spatial_dim=input_spatial_dim,
        targets_spatial_dim=targets_spatial_dim,
        blank_index=blank_index,
        max_approx=max_approx,
    )


def edit_distance(a: Tensor, a_spatial_dim: Dim, b: Tensor, b_spatial_dim: Dim, *, dtype: str = "int32") -> Tensor:
    """
    :param a: [B,Ta]
    :param a_spatial_dim: Ta
    :param b: [B,Tb]
    :param b_spatial_dim: Tb
    :param dtype:
    :return: [B]
    """
    import numpy  # just for iinfo on dtype to get max value

    # The axis permutation is just an efficiency optimization.
    a = a.copy_transpose([a_spatial_dim] + a.remaining_dims(a_spatial_dim))
    b = b.copy_transpose([b_spatial_dim] + b.remaining_dims(b_spatial_dim))
    dev = a.device
    max_dist_err = numpy.iinfo(dtype).max
    n_a_max_len = a_spatial_dim.get_dim_value()
    n_b_max_len = b_spatial_dim.get_dim_value()
    if int(n_a_max_len) < int(n_b_max_len):
        a, b = b, a
        a_spatial_dim, b_spatial_dim = b_spatial_dim, a_spatial_dim
        n_a_max_len, n_b_max_len = n_b_max_len, n_a_max_len
    # Now n_a_max_len >= n_b_max_len.
    batch_dims = a.remaining_dims(a_spatial_dim)
    for dim in b.remaining_dims(b_spatial_dim):
        if dim not in batch_dims:
            batch_dims.append(dim)
    a_seq_len = a_spatial_dim.get_dyn_size_ext_for_device(dev)  # [B]
    b_seq_len = b_spatial_dim.get_dyn_size_ext_for_device(dev)  # [B]
    a_tensor_ext, (a_spatial_dim_ext,) = rf.pad(
        a, axes=[a_spatial_dim], padding=[(b_spatial_dim, b_spatial_dim)], handle_dynamic_dims=False
    )  # [Tb+Ta+Tb,B]
    a_spatial_dim_ext: Dim
    b_tensor_flipped = rf.reverse_sequence(b, axis=b_spatial_dim, handle_dynamic_dims=False)  # [Tb,B]
    entry_idx_ = rf.range_over_dim(b_spatial_dim, device=dev)  # [Tb]->Tb
    b_spatial_dim1 = b_spatial_dim + 1
    buffer_dim = Dim(3 * b_spatial_dim1.get_dim_value_tensor(), name="buffer")
    buffer = rf.Parameter([buffer_dim] + batch_dims, device=dev, dtype=dtype, auxiliary=True)  # [3*(Tb+1),B]
    buffer_offsets = [0, b_spatial_dim1.get_dim_value_tensor(), b_spatial_dim1.get_dim_value_tensor() * 2]
    result = rf.where((a_seq_len == 0) & (b_seq_len == 0), 0, max_dist_err)  # [B]  # noqa

    # We are going diagonal over (Ta+1) and (Tb+1). (Similar as RETURNN native EditDistanceOp.)
    # You need to draw the grid on paper to understand all the index math...
    for u in range(1, n_a_max_len + n_b_max_len + 1):

        prev2_dist, _ = rf.slice(
            buffer, axis=buffer_dim, start=buffer_offsets[u % 3], size=b_spatial_dim1, out_dim=b_spatial_dim1
        )  # [Tb+1,B]
        prev_dist, _ = rf.slice(
            buffer, axis=buffer_dim, start=buffer_offsets[(u + 1) % 3], size=b_spatial_dim1, out_dim=b_spatial_dim1
        )  # [Tb+1,B]
        cur_dist_start_offset = buffer_offsets[(u + 2) % 3]

        del_cost = (
            rf.slice(prev_dist, axis=b_spatial_dim1, end=b_spatial_dim.get_dim_value_tensor(), out_dim=b_spatial_dim)[0]
            + 1
        )  # [Tb,B]
        ins_cost = rf.slice(prev_dist, axis=b_spatial_dim1, start=1, out_dim=b_spatial_dim)[0] + 1  # [Tb,B]
        sub_cost = rf.slice(prev2_dist, axis=b_spatial_dim1, start=1, out_dim=b_spatial_dim)[0] + rf.cast(
            rf.slice(a_tensor_ext, axis=a_spatial_dim_ext, start=u - 1, size=b_spatial_dim, out_dim=b_spatial_dim)[0]
            != b_tensor_flipped,
            dtype=dtype,
        )
        min_cost = rf.minimum(del_cost, ins_cost, sub_cost)  # [Tb,B]
        t_a_gt_zero_mask = entry_idx_ > n_b_max_len - u  # [Tb]

        buffer.assign_key(
            axis=buffer_dim,
            key=slice(cur_dist_start_offset, cur_dist_start_offset + b_spatial_dim.get_dim_value_tensor()),
            key_dim=b_spatial_dim,
            value=rf.where(t_a_gt_zero_mask, min_cost, u),
        )
        # last entry in cur_dist, that is where t_b == 0
        buffer.assign_key(
            axis=buffer_dim, key=cur_dist_start_offset + b_spatial_dim.get_dim_value_tensor(), key_dim=None, value=u
        )

        end_offset_a = n_b_max_len + a_seq_len - u  # [B]
        end_offset_b = n_b_max_len - b_seq_len  # [B]
        result = rf.where(
            end_offset_a == end_offset_b,
            rf.gather(buffer, axis=buffer_dim, indices=cur_dist_start_offset + end_offset_a, clip_to_valid=True),
            result,
        )

    return result



"""
Functions to get the network dict based on the config
via the API get_model, train_step, forward_step.

https://github.com/rwth-i6/returnn/issues/1120
"""

from __future__ import annotations
from typing import Any, Dict, Tuple
import tensorflow as tf
from tensorflow.python.util import nest
import returnn.tf.compat as tf_compat
from returnn.util.basic import BehaviorVersion
from returnn.tensor import TensorDict, Tensor, Dim, batch_dim
from returnn.config import get_global_config
from returnn.tf.network import LayerBase
import returnn.frontend as rf
from .. import frontend_layers as rfl
from . import _utils


__all__ = ["get_net_dict"]


def get_net_dict(
    *,
    epoch: int,
    step: int,
) -> Tuple[Dict[str, Any], rf.Module]:
    """called from the RETURNN config"""
    BehaviorVersion.set_min_behavior_version(rfl.min_returnn_behavior_version)
    rf.select_backend_returnn_layers_tf()
    rfl.Layer.reset_default_root()

    config = get_global_config()

    # See :mod:`rf.rand` docstring for an explanation of this logic.
    random_seed = config.int("random_seed", 42)
    random_seed = (epoch * 193939 + step * 19937 + random_seed * 27644437 + 479001599) % (2**31)
    rf.set_random_seed(random_seed)

    extern_data = TensorDict()
    extern_data_dict = config.typed_value("extern_data")
    extern_data.update(extern_data_dict, auto_convert=True)
    assert isinstance(extern_data, TensorDict)
    for data in extern_data.data.values():
        rfl.register_extern_data(data)

    get_model_func = config.typed_value("get_model")
    model = get_model_func(epoch=epoch, step=step)

    task = config.value("task", None)
    step_tensor = rfl.make_layer({"class": "global_train_step"}, name="global_train_step")
    epoch_tensor = rfl.make_layer(
        {"class": "eval", "eval": _eval_func_get_epoch, "from": (), "out_type": {"dims": (), "dtype": "int32"}},
        name="epoch",
    )
    if task in {"train", "eval"}:
        rf.init_train_step_run_ctx(
            train_flag=rfl.make_layer({"class": "train_flag"}, name="train_flag"),
            step=step_tensor,
            epoch=epoch_tensor,
        )
        train_step_func = get_global_config().typed_value("train_step")
        train_step_func(
            model=model,
            extern_data=extern_data,
        )
    elif task in {"forward", "search"}:
        rf.init_forward_step_run_ctx(step=step_tensor, epoch=epoch_tensor)
        forward_step_func = get_global_config().typed_value("forward_step")
        forward_step_func(
            model=model,
            extern_data=extern_data,
        )
    else:
        raise ValueError(f"invalid task {task!r}")

    root_scope = rfl.Layer.top().root

    for loss in rf.get_run_ctx().losses.values():
        loss_t = _utils.copy(loss.loss, name=root_scope.get_new_child(suggested_name=loss.name))
        loss_t.raw_tensor.layer_dict["loss"] = "as_is"
        loss_opts = {}
        if loss.scale != 1:
            assert "loss_scale" not in loss_t.raw_tensor.layer_dict
            loss_opts["scale"] = loss.scale
        if loss.as_error:
            loss_opts["as_error"] = True
        if loss.use_normalized_loss:
            loss_opts["use_normalized_loss"] = True
        if not loss.use_flatten_frames:
            loss_opts["use_flatten_frames"] = False
        if loss.custom_inv_norm_factor is not None:
            loss_opts["custom_inv_norm_factor"] = loss.custom_inv_norm_factor
        if loss_opts:
            loss_t.raw_tensor.layer_dict["loss_opts"] = loss_opts
        # Add it to the root name scope marked_losses list.
        # Note that this logic might change.
        root_scope.marked_losses.append(loss_t)

    for out in rf.get_run_ctx().outputs.data.values():
        if out.name == "output" and out.name not in root_scope.children:
            layer = root_scope.get_child(out.name)
        else:
            layer = root_scope.get_new_child(suggested_name=out.name)
        out_t = _utils.copy(out, name=layer)
        if layer.name != "output":
            out_t.raw_tensor.layer_dict["is_output_layer"] = True
        root_scope.marked_outputs.append(out_t)

    net_dict = root_scope.get_returnn_config(root_module=model).get_net_dict_raw_dict()

    def _cleanup_net_dict_value(elem):
        assert not isinstance(elem, Tensor), f"not expected to see Tensor {elem} in net dict"
        if isinstance(elem, Dim):
            # Dim dyn_size_ext might be Tensor[rfl.Layer],
            # but now the TF engine actually wants to have Tensor[tf.Tensor].
            # Reset it now. The TF engine should redefine it again.
            elem.reset_batch_and_raw()
        elif isinstance(elem, set):
            # map_structure does not recurse into sets.
            nest.map_structure(_cleanup_net_dict_value, sorted(list(elem)))
        return elem

    # Do some cleanup.
    nest.map_structure(_cleanup_net_dict_value, net_dict)
    _cleanup_net_dict_value(batch_dim)
    for data in extern_data.data.values():
        for dim in data.dims:
            _cleanup_net_dict_value(dim)
    return net_dict, model


def _eval_func_get_epoch(self: LayerBase, **_kwargs) -> tf.Tensor:
    # Currently this is almost never used, so we don't introduce an own custom layer for this,
    # nor any TF placeholder or TF variable, but just use this.
    # A TF variable might be the next simplest solution, if we want to improve this later...
    # See TFNetwork.set_run_opts().
    run_opts = self.network.get_root_network().get_run_opts()

    def _py_func_get_epoch() -> int:
        return run_opts["epoch"]

    (epoch,) = tf_compat.v1.py_func(_py_func_get_epoch, [], [tf.int32], stateful=True)
    assert isinstance(epoch, tf.Tensor)
    epoch.set_shape(())
    return epoch



"""
torch.distributed utils
"""

from __future__ import annotations
from typing import Optional, Any, Dict
import os
import socket
import logging

import torch
from torch.nn.parallel import DistributedDataParallel

from returnn.config import Config
from returnn.util.basic import CollectionReadCheckCovered

_logger = logging.getLogger("returnn.torch.distributed")


class DistributedContext:
    """
    This class setups some helper functions for torch distributed training
    """

    def __init__(self, options: Dict[str, Any]):
        import torch.distributed as dist

        self._opts = CollectionReadCheckCovered(options)

        # when no backend is specified, both gloo and nccl backends will be created
        # the gloo backend will be used for collectives with CPU tensors and
        # the nccl backend will be used for collectives with CUDA tensors
        dist.init_process_group(backend=self._opts.get("backend", None))

        self._local_rank = int(os.environ["LOCAL_RANK"])
        self._local_size = int(os.environ["LOCAL_WORLD_SIZE"])
        self._rank = dist.get_rank()
        self._size = dist.get_world_size()

        _logger.info(
            "Torch distributed initialized. Hostname %s, pid %i, rank %i / size %i, local rank %s / local size %s."
            % (socket.gethostname(), os.getpid(), self._rank, self._size, self._local_rank, self._local_size)
        )

        self._reduce_type = self._opts.get("reduce_type", "grad")
        self._param_sync_step: Optional[int] = self._opts.get("param_sync_step", None)
        if self._reduce_type == "param":
            assert isinstance(self._param_sync_step, int) and self._param_sync_step > 0, (
                f"reduce_type param: param_sync_step must be a positive int,"
                f" got {self._param_sync_step!r} ({type(self._param_sync_step).__name__})"
            )
            _logger.info(f"reduce_type param: param_sync_step {self._param_sync_step}")
        elif self._reduce_type == "grad":
            _logger.info("reduce_type grad")
        else:
            raise ValueError(f"invalid reduce_type {self._reduce_type!r}")

        self._check_no_unknown_opts()

    def __repr__(self):
        return f"<{self.__class__.__name__} size={self._size} reduce_type={self._reduce_type}>"

    def _check_no_unknown_opts(self):
        # We check that all opts in self._opts have been used.
        # This function here is called at the end in __init__,
        # and not all opts are used yet, so read them now,
        # such that the check in the end works.
        if self._reduce_type == "grad":
            self._opts.get("class")
            self._opts.get("options")
        if self._reduce_type == "param":
            self._opts.get("sync_on_cpu")

        self._opts.assert_all_read()

    def local_rank(self) -> int:
        """local rank"""
        return self._local_rank

    def local_size(self) -> int:
        """local size"""
        return self._local_size

    def rank(self) -> int:
        """global rank"""
        return self._rank

    def size(self) -> int:
        """global size"""
        return self._size

    def get_param_sync_step(self) -> Optional[int]:
        """param sync step"""
        return self._param_sync_step

    def maybe_make_distributed_module(self, module: torch.nn.Module) -> Optional[DistributedDataParallel]:
        """
        Maybe make a wrapped distributed module.

        :param module: original module
        :return: potentially wrapped module
        """
        if self._reduce_type == "param":
            return None
        assert self._reduce_type == "grad"
        cls = self._opts.get("class", DistributedDataParallel)
        if cls is not DistributedDataParallel:
            _logger.warning(f"Using custom class {cls} instead of DistributedDataParallel, might be unsupported.")
        kwargs = self._opts.get("options", {})
        return cls(
            module=module,
            device_ids=[self.local_rank()],
            **kwargs,
        )

    def step_after_param_update(self, *, module: torch.nn.Module, epoch_step_idx: int):
        """one train step"""
        if self._reduce_type == "param" and ((epoch_step_idx % self._param_sync_step) == (self._param_sync_step - 1)):
            _sync_params_avg(module=module, sync_on_cpu=self._opts.get("sync_on_cpu", False))


_is_set_up = False
_ctx = None  # type: Optional[DistributedContext]


def get_ctx(config=None) -> Optional[DistributedContext]:
    """
    :param Config|None config:
    :returns: the global context if Torch distributed is enabled, or None otherwise.
      If we did not setup the context yet, it will automatically create it.
    """
    global _is_set_up, _ctx
    if _is_set_up:
        return _ctx
    if not config:
        from returnn.config import get_global_config

        config = get_global_config(raise_exception=False)
        if not config:
            return None

    _is_set_up = True
    opts = config.typed_value("torch_distributed")
    if opts is None:
        return None

    assert isinstance(opts, dict)
    _ctx = DistributedContext(opts)
    return _ctx


@torch.no_grad()
def _sync_params_avg(*, module: torch.nn.Module, sync_on_cpu: bool = False):
    import torch.distributed as dist

    if sync_on_cpu:
        for param in module.parameters():
            # Separately move each param to CPU (instead of the whole module), to safe CPU memory.
            param_cpu = param.to(torch.device("cpu"))
            # On CPU, we are likely using Gloo, and Gloo does not support AVG
            dist.all_reduce(param_cpu.data, op=dist.ReduceOp.SUM)
            param_cpu.data /= dist.get_world_size()
            param.data = param_cpu.to(param.device)
        return

    if dist.get_backend() == "gloo":
        # Gloo does not support AVG
        reduce_op = dist.ReduceOp.SUM
    else:
        if hasattr(dist.ReduceOp, "AVG"):
            reduce_op = dist.ReduceOp.AVG
        else:
            # Older PyTorch versions do not have ReduceOp.AVG.
            reduce_op = dist.ReduceOp.SUM

    for param in module.parameters():
        dist.all_reduce(param.data, op=reduce_op)
        if reduce_op == dist.ReduceOp.SUM:
            param.data /= dist.get_world_size()



"""
High-level backend for RETURNN layers
"""

from __future__ import annotations
from typing import TypeVar, Union, Sequence, Optional, Any, Callable, Tuple, Dict, List
import contextlib
import numpy
import tensorflow as tf
import returnn.tf.compat as tf_compat

from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim

# noinspection PyProtectedMember
from returnn.frontend._backend import Backend

from .layer import Layer
from .. import frontend_layers as rfl
from ... import frontend as rf
from . import dims as _dims
from returnn.frontend import RawTensorTypes

# noinspection PyProtectedMember
from returnn.frontend import _random_journal


S = TypeVar("S")  # any nested structure, can be None


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class ReturnnLayersBackend(Backend[Layer]):
    """
    RETURNN layers backend (using TF), where raw_tensor represents a RETURNN layer
    """

    name = "returnn_layers_tf"
    RawTensorType = Layer
    is_tensorflow = True
    is_backend_raw_tensor_dim_tag_independent = False

    @staticmethod
    def executing_eagerly() -> bool:
        """executing eagerly"""
        return False

    @staticmethod
    def get_tensor_dependencies(x: Tensor[Layer]) -> Sequence[Tensor]:
        """get tensor inputs"""
        deps: List[Tensor] = []
        visited = set()
        for dep in x.raw_tensor.get_tensor_dependencies():
            if not dep.tensor or dep.tensor in visited:
                continue
            visited.add(dep.tensor)
            deps.append(dep.tensor)
        return deps

    @staticmethod
    def get_tensor_consumers(x: Tensor[Layer]) -> Sequence[Tensor]:
        """get tensor consumers"""
        usages: List[Tensor] = []
        visited = set()
        for use in x.raw_tensor.usages:
            if not use.tensor or use.tensor in visited:
                continue
            visited.add(use.tensor)
            usages.append(use.tensor)
        return usages

    @staticmethod
    def cond(pred: Tensor, true_fn: Callable, false_fn: Callable):
        """cond"""
        with rfl.Cond(pred) as cond:
            cond.true = true_fn()
            cond.false = false_fn()
        return cond.result

    @staticmethod
    def while_loop(
        cond: Callable[[S], Union[bool, Tensor]],
        body: Callable[[S], S],
        initial: S,
    ) -> S:
        """while loop"""

        # Have to put some arbitrary max_seq_len limit, otherwise the RecLayer will complain.
        # Note that we could put the `loop.end(..., include_eos=False)` at the beginning of the loop.
        # However, then all layers run one more iteration than you might expect,
        # specifically the last iteration will be run with end being True (condition being False).
        # There are multiple cases when this can be problematic:
        #  - Layers with side-effects.
        #  - Gathering or similar where the indices can get out of bounds.
        # When we have `loop.end(..., include_eos=True)` at the end of the loop, this is fixed.
        # However, the case where the `cond(...)` is False already in the very beginning is wrong now.
        # To check for this, we have to add an additional `cond(...)` around the loop.
        def _loop():
            loop = rfl.Loop(max_seq_len=rf.constant(2**31 - 1, dims=()))
            loop.state.state = initial
            with loop:
                loop.state.state = body(loop.state.state)
                loop.end(rf.logical_not(cond(loop.state.state)), include_eos=True)
            return loop.state.state

        return ReturnnLayersBackend.cond(cond(initial), _loop, lambda: initial)

    @staticmethod
    def set_random_seed(seed: int):
        """
        :param seed:
        """
        tf_compat.v1.set_random_seed(seed)

    @staticmethod
    def get_random_state() -> Dict[str, bytes]:
        """
        :return: random state
        """
        # Not sure... we could cover all state vars used by RandomLayer...
        return {}

    @staticmethod
    def set_random_state(state: Dict[str, bytes]):
        """
        :param state: as returned by :func:`get_random_state`.
            This might not always be successful (e.g. different hardware, different backend version),
            so the calling code should always have called set_random_seed before to have the random generators
            in a reasonable fallback state.
        """
        assert not state  # not implemented... unexpected currently

    @staticmethod
    def get_dtype_name_raw(raw_tensor: Layer) -> str:
        """dtype"""
        return raw_tensor.tensor.dtype

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> str:
        """dtype"""
        return dtype_name

    @staticmethod
    def get_ndim_raw(raw_tensor: Layer) -> int:
        """ndim"""
        return raw_tensor.tensor.batch_ndim

    @staticmethod
    def get_shape_raw(raw_tensor: Layer) -> Layer:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: Layer) -> Tuple[Union[int, Layer], ...]:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: Layer) -> Tuple[Optional[int], ...]:
        """known shape"""
        return raw_tensor.tensor.batch_shape

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, Layer]], Layer], value: Union[Any, Layer]) -> Layer:
        """fill raw"""
        raise Exception("fill_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def compare_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """compare"""
        raise Exception("compare_raw should not get called")

    @staticmethod
    def combine_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """combine"""
        raise Exception("combine_raw should not get called")

    @staticmethod
    def reshape_raw(raw_tensor: Layer, shape: Union[Sequence[Union[int, Layer]], Layer]) -> Layer:
        """reshape_raw"""
        raise Exception("reshape_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def make_output_tensor(tensor: Tensor, dims: Sequence[Dim], *, name: str) -> Tensor:
        """only func where we have explicitly defined dim order in the output"""
        return rfl.make_layer({"class": "transpose", "from": tensor, "perm": dims}, name=name)

    @staticmethod
    def copy(tensor: Tensor) -> Tensor:
        """copy"""
        return rfl.make_layer({"class": "identity", "from": tensor}, name="copy")

    @staticmethod
    def cast(tensor: Tensor, dtype: str) -> Tensor:
        """cast"""
        return rfl.make_layer({"class": "cast", "from": tensor, "dtype": dtype}, name="cast")

    @staticmethod
    def set_requires_gradient(tensor: Tensor):
        """
        set requires gradient; not needed for TensorFlow, will always calculate whatever is needed
        """

    @staticmethod
    def gradient(y: Tensor, x: Tensor) -> Tensor:
        """gradient"""
        return rfl.make_layer({"class": "gradient", "y": y, "x": x}, name="gradient")

    @staticmethod
    def stop_gradient(tensor: Tensor) -> Tensor:
        """stop grad"""
        return rfl.make_layer({"class": "scaled_grad", "from": tensor, "scale": 0}, name="stop_gradient")

    @staticmethod
    def scaled_gradient(tensor: Tensor, scale: Union[float, Tensor]) -> Tensor:
        """scaled gradient"""
        return rfl.make_layer({"class": "scaled_grad", "from": tensor, "scale": scale}, name="scaled_gradient")

    @staticmethod
    def scaled_gradient_ext(
        x: Tensor,
        *,
        scale: Union[float, Tensor] = 1.0,
        shift: Optional[Union[float, Tensor]] = None,
        scale_shift_by_sum_over_axis: Optional[Dim] = None,
    ):
        """scaled gradient ext"""
        return rfl.make_layer(
            {
                "class": "scaled_grad",
                "from": x,
                "scale": scale,
                "shift": shift,
                "scale_shift_by_sum_over_axis": scale_shift_by_sum_over_axis,
            },
            name="scaled_gradient_ext",
        )

    @staticmethod
    def merge_dims(
        source: Tensor,
        *,
        dims: Sequence[Dim],
        out_dim: Optional[Dim] = None,
    ) -> Tuple[Tensor, Dim]:
        """
        Merges a list of axes into a single one. (Flatten the dims.)
        E.g. input is (batch, width, height, dim) and dims=(width,height), then we get (batch, width*height, dim).
        Or input is (batch, time, height, dim) and axes=(height,dim), then we get (batch, time, height*dim).

        :param source:
        :param dims:
        :param out_dim:
        :return: tensor, out_dim
        """
        if not isinstance(source, Tensor):
            raise TypeError(f"merge_dims: unexpected type for source {source!r}, need tensor")
        if out_dim is None:
            out_dim = dims[0]
            for d in dims[1:]:
                out_dim = out_dim * d
        layer = rfl.make_layer(
            {"class": "merge_dims", "from": source, "axes": dims, "out_dim": out_dim}, name="merge_dims"
        )
        return layer, out_dim

    @staticmethod
    def split_dims(
        source: Tensor,
        *,
        axis: Dim,
        dims: Sequence[Dim],
        pad_to_multiples: Optional[bool] = None,
        pad_value: Union[None, int, float] = None,
    ) -> Tensor:
        """split dims"""
        args = {}
        if pad_to_multiples is not None or pad_value is not None:
            args["pad_to_multiples"] = pad_to_multiples
            args["pad_value"] = pad_value
        args = {key: value for (key, value) in args.items() if value is not NotSpecified}
        return rfl.make_layer(
            {"class": "split_dims", "from": source, "axis": axis, "dims": dims, **args}, name="split_dims"
        )

    @staticmethod
    def reshape(source: Tensor, in_dims: Sequence[Dim], out_dims: Sequence[Dim]) -> Tensor:
        """reshape"""
        return rfl.make_layer(
            {
                "class": "reshape",
                "from": source,
                "in_dims": in_dims,
                "out_dims": out_dims,
                "extra_deps": rfl.get_dim_deps(out_dims),
            },
            name="reshape",
        )

    @staticmethod
    def split(source: Tensor, *, axis: Dim, out_dims: Sequence[Dim]) -> Tuple[Tensor, ...]:
        """split"""
        res = rfl.make_layer({"class": "split", "from": source, "axis": axis, "out_dims": out_dims}, name="split")

        src_axis_int = source.get_axis_from_description(axis)
        # noinspection PyProtectedMember
        return tuple(
            rfl._get_sub_layer(
                layer=res,
                name=str(i),
                data=source.copy_template_replace_dim_tag(
                    axis=src_axis_int, new_dim_tag=dim, name=f"{source.name}/split:{i}:{dim.description}"
                ),
            )
            for i, dim in enumerate(out_dims)
        )

    @staticmethod
    def expand_dim(source: Tensor, dim: Dim) -> Tensor:
        """expand dim"""
        # Some heuristic on the kind, which just determines where RETURNN puts the new axis.
        if source.have_feature_axis():
            axis = "spatial"
        elif dim.dimension is not None:
            axis = "feature"
        else:
            axis = "spatial"
        return rfl.make_layer({"class": "expand_dims", "from": source, "axis": axis, "dim": dim}, name="expand_dims")

    @staticmethod
    def squeeze(source: Tensor, axis: Dim) -> Tensor:
        """squeeze"""
        return rfl.make_layer({"class": "squeeze", "from": source, "axis": axis}, name="squeeze")

    @staticmethod
    def concat(
        *sources: Tuple[Tensor, Dim],
        allow_broadcast: bool = False,
        out_dim: Dim,
    ) -> Tensor:
        """concat"""
        opts = {}
        if allow_broadcast:
            opts["allow_broadcast"] = True
        dim_deps = rfl.get_dim_deps(out_dim)
        sources_dims = set()
        for source, _ in sources:
            sources_dims.update(source.dims)
        need_explicit_dim_deps = False
        for dim in dim_deps:
            if dim not in sources_dims:
                need_explicit_dim_deps = True
                break
        if need_explicit_dim_deps:
            source0 = rfl.make_layer(
                {"class": "copy", "from": sources[0][0], "extra_deps": dim_deps}, name="concat_extra_dim_deps"
            )
            sources = ((source0, sources[0][1]),) + sources[1:]
        return rfl.make_layer(
            {"class": "concat", "from": sources, "out_dim": out_dim, **opts},
            name="concat",
        )

    @staticmethod
    def pad(
        source: Tensor,
        *,
        axes: Sequence[Dim],
        padding: Sequence[Tuple[Union[Dim, int, Tensor], Union[Dim, int, Tensor]]],
        out_dims: Sequence[Dim],
        handle_dynamic_dims: bool,
        mode: str = "constant",
        value: Union[rf.RawTensorTypes, Tensor] = None,
    ) -> Tensor:
        """pad"""
        assert isinstance(value, (int, float, type(None)))  # not implemented otherwise
        return rfl.make_layer(
            {
                "class": "pad",
                "from": source,
                "axes": axes,
                "padding": padding,
                "out_dims": out_dims,
                "handle_dynamic_dims": handle_dynamic_dims,
                "mode": mode,
                "value": value,
            },
            name="pad",
        )

    @staticmethod
    def activation(tensor: Tensor, func: str) -> Tensor:
        """activation"""
        return rfl.make_layer({"class": "activation", "activation": func, "from": tensor}, name=func)

    @staticmethod
    def activation_raw(raw_tensor: Layer, func: str) -> Layer:
        """activation"""
        return rfl.make_layer(
            {"class": "activation", "activation": func, "from": raw_tensor.tensor}, name=func
        ).raw_tensor

    @staticmethod
    def safe_log(tensor: Tensor, *, eps: float) -> Tensor:
        """safe log"""
        return rfl.make_layer(
            {"class": "activation", "activation": "safe_log", "eps": eps, "from": tensor}, name="safe_log"
        )

    @staticmethod
    def softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """softmax"""
        args = {}
        if not use_mask:
            args["use_time_mask"] = False
        return rfl.make_layer({"class": "softmax_over_spatial", "axis": axis, "from": tensor, **args}, name="softmax")

    @staticmethod
    def log_softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """log softmax"""
        args = {}
        if not use_mask:
            args["use_time_mask"] = False
        return rfl.make_layer(
            {"class": "softmax_over_spatial", "axis": axis, "from": tensor, "log_space": True, **args},
            name="log_softmax",
        )

    @staticmethod
    def softmax_cross_entropy_with_logits(*, logits: Tensor, targets: Tensor, axis: Dim):
        """
        Efficient cross entropy.

        :param logits: target estimates given as inputs to softmax (i.e. unnormalized)
        :param targets: probabilities, i.e. normalized, can also be sparse
        :param axis: class labels dim over which softmax is computed
        :return: cross entropy (same Dims as 'logits' but without 'axis')
        """
        if targets.sparse_dim:
            assert axis == targets.sparse_dim and axis in logits.dims
            return rfl.make_layer(
                {
                    "class": "sparse_softmax_cross_entropy_with_logits",
                    "logits": logits,
                    "targets": targets,
                    "axis": axis,
                },
                name="sparse_softmax_cross_entropy_with_logits",
            )
        else:  # dense targets
            assert axis in targets.dims and axis in logits.dims
            log_probs = rf.log_softmax(logits, axis=axis)
            return -rf.matmul(targets, log_probs, reduce=axis)

    @staticmethod
    def ctc_loss(
        *,
        logits: Tensor,
        logits_normalized: bool = False,
        targets: Tensor,
        input_spatial_dim: Dim,
        targets_spatial_dim: Dim,
        blank_index: int,
        max_approx: bool = False,
    ) -> Tensor:
        """CTC"""
        assert targets.sparse_dim and targets.sparse_dim.dimension <= logits.feature_dim.dimension
        logits = rfl.make_layer(
            {"class": "reinterpret_data", "from": logits, "set_axes": {"T": input_spatial_dim}}, name="logits"
        )
        targets = rfl.make_layer(
            {"class": "reinterpret_data", "from": targets, "set_axes": {"T": targets_spatial_dim}}, name="targets"
        )
        return rfl.make_layer(
            {
                "class": "ctc_loss",
                "logits": logits,
                "logits_normalized": logits_normalized,
                "targets": targets,
                "blank_index": blank_index,
                "max_approx": max_approx,
            },
            name="ctc_loss",
        )

    @staticmethod
    def create_parameter_raw(tensor: rf.Parameter, *, device: Optional[str] = None) -> Layer:
        """create parameter"""
        return rfl.make_layer(
            {
                "class": "variable",
                "shape": tensor.dims,
                "dtype": tensor.dtype,
                "param_name": "param",
                "param_device": device,
            },
            name=tensor.name,
            out=tensor,
        ).raw_tensor

    @staticmethod
    def set_parameter_initial_value(param: rf.Parameter[Layer], value: Union[None, Tensor, rf.RawTensorTypes]) -> None:
        """set parameter initial value"""
        if value is None:
            param.raw_tensor.layer_dict.pop("init", None)
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
        elif isinstance(value, Tensor):
            param.raw_tensor.layer_dict.pop("init", None)
            if not value.raw_tensor.parent.can_access_children_from_root:
                accessible_parent = value.raw_tensor.parent
                while not accessible_parent.can_access_children_from_root:
                    accessible_parent = accessible_parent.parent
                value.raw_tensor.assign_parent(accessible_parent)
                # We could also maybe move out all the dependencies.
                # However, it's not clear whether this is always safe.
                for dep in value.raw_tensor.get_tensor_dependencies():
                    assert (
                        dep.parent.can_access_children_from_root
                    ), f"dep {dep} of moved value {value} is not accessible"
            param.raw_tensor.layer_dict["init_by_layer"] = value
        else:
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
            param.raw_tensor.layer_dict["init"] = value
        if rfl.is_debug_eager_mode_enabled():
            shape = [d.get_dim_value() for d in param.dims]
            if isinstance(value, Tensor):
                assert value.placeholder is not None
                value_tf = value.placeholder
            else:
                value_tf = tf.broadcast_to(tf.convert_to_tensor(value), shape)
            if param.raw_tensor.debug_layer.output.placeholder is None:
                var = tf.Variable(value_tf, shape=[d.get_dim_value() for d in param.dims], dtype=param.dtype)
                param.raw_tensor.debug_layer.output.placeholder = var
            else:
                var = param.raw_tensor.debug_layer.output.placeholder
                assert isinstance(var, tf.Variable)
                var.assign(value_tf)

    @staticmethod
    def set_parameter_trainable(param: rf.Parameter, trainable: bool) -> None:
        """set parameter trainable"""
        if trainable:
            # pop it, as it's the default
            param.raw_tensor.layer_dict.pop("trainable", None)
        else:
            param.raw_tensor.layer_dict["trainable"] = False

    @staticmethod
    def parameter_assign(param: rf.Parameter, value: Tensor, *, op: str = "assign") -> None:
        """param assign"""
        from .parameter_assign import parameter_assign

        parameter_assign(param=param, value=value, op=op)

    @staticmethod
    def convert_to_tensor(
        value: Union[Tensor, Layer, RawTensorTypes],
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        device: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Tensor[Layer]:
        """convert to tensor"""
        if isinstance(value, Tensor):
            return value
        kwargs = {}
        if sparse_dim:
            kwargs["sparse_dim"] = sparse_dim
        dim_deps = _dims.get_dim_deps(dims)
        if dim_deps:
            kwargs["shape_deps"] = dim_deps
        return rfl.make_layer(
            {"class": "constant", "value": value, "shape": dims, "dtype": dtype, **kwargs},
            name=name or "convert_to_tensor",
        )

    @staticmethod
    def full(
        dims: Sequence[Dim],
        fill_value: Union[RawTensorTypes, Tensor],
        *,
        dtype: str,
        device: Optional[str] = None,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
    ) -> Tensor:
        """full"""
        if isinstance(fill_value, Tensor):
            return fill_value + rf.zeros_like(fill_value)
        kwargs = {}
        if sparse_dim:
            kwargs["sparse_dim"] = sparse_dim
        kwargs["feature_dim"] = feature_dim
        dim_deps = _dims.get_dim_deps(dims)
        if dim_deps:
            kwargs["shape_deps"] = dim_deps
        return rfl.make_layer(
            {"class": "constant", "value": fill_value, "shape": dims, "dtype": dtype, **kwargs}, name="full"
        )

    @classmethod
    def compare(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """compare"""
        kwargs = {}
        if allow_broadcast_all_sources is not None:
            kwargs["allow_broadcast_all_sources"] = allow_broadcast_all_sources
        a = rf.convert_to_tensor(a, _backend=cls)
        b = rf.convert_to_tensor(b, _backend=cls)
        return rfl.make_layer({"class": "compare", "from": [a, b], "kind": kind, **kwargs}, name=kind)

    @classmethod
    def combine(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """combine"""
        kwargs = {}
        if allow_broadcast_all_sources is not None:
            kwargs["allow_broadcast_all_sources"] = allow_broadcast_all_sources
        a = rf.convert_to_tensor(a, _backend=cls)
        b = rf.convert_to_tensor(b, _backend=cls)
        return rfl.make_layer({"class": "combine", "from": [a, b], "kind": kind, **kwargs}, name=kind)

    @staticmethod
    def gather(source: Tensor, *, indices: Union[Tensor, int], axis: Dim, clip_to_valid: bool = False) -> Tensor:
        """gather"""
        args = {}
        if clip_to_valid:
            args["clip_to_valid"] = clip_to_valid
        return rfl.make_layer(
            {"class": "gather", "from": source, "position": indices, "axis": axis, **args}, name="gather"
        )

    @staticmethod
    def slice(
        source: Tensor,
        *,
        axis: Dim,
        start: Optional[Union[int, Tensor]] = None,
        end: Optional[Union[int, Tensor]] = None,
        step: Optional[Union[int, Tensor]] = None,
        size: Optional[Union[int, Tensor, Dim]] = None,
        out_dim: Dim,
    ) -> Tensor:
        """slice"""
        if size is not None:
            assert end is None  # not implemented
            assert step is None  # not implemented
            assert size is not None  # not implemented
            return rfl.make_layer(
                {
                    "class": "slice_nd",
                    "from": source,
                    "axis": axis,
                    "start": start,
                    "size": size,
                    "out_spatial_dim": out_dim,
                },
                name="slice_nd",
            )
        assert size is None  # not implemented
        assert isinstance(start, (int, type(None)))  # not implemented
        assert isinstance(end, (int, type(None)))  # not implemented
        assert isinstance(step, (int, type(None)))  # not implemented
        args = {}
        if start is not None:
            args["slice_start"] = start
        if end is not None:
            args["slice_end"] = end
        if step is not None:
            args["slice_step"] = step
        return rfl.make_layer(
            {"class": "slice", "from": source, "axis": axis, "out_dim": out_dim, **args}, name="slice"
        )

    @staticmethod
    def flip(source: Tensor, *, axis: Dim) -> Tensor:
        """flip"""
        return rfl.make_layer(
            {"class": "slice", "from": source, "axis": axis, "out_dim": axis, "slice_step": -1}, name="flip"
        )

    @staticmethod
    def where(
        cond: Tensor,
        true_: Union[Tensor, rf.RawTensorTypes],
        false_: Union[Tensor, rf.RawTensorTypes],
        *,
        allow_broadcast_all_sources: bool = False,
    ) -> Tensor:
        """where"""
        allow_broadcast_all_sources  # noqa # ignore allow_broadcast_all_sources for now..., not implemented
        return rfl.make_layer(
            {"class": "switch", "condition": cond, "true_from": true_, "false_from": false_}, name="where"
        )

    @staticmethod
    def clip_by_value(
        x: Tensor,
        clip_value_min: Union[Tensor, rf.RawTensorTypes],
        clip_value_max: Union[Tensor, rf.RawTensorTypes],
        *,
        allow_broadcast_all_sources: bool = False,
    ) -> Tensor:
        """clip by value"""
        clip_value_min = rf.convert_to_tensor(clip_value_min, _backend=ReturnnLayersBackend)
        clip_value_max = rf.convert_to_tensor(clip_value_max, _backend=ReturnnLayersBackend)
        return rfl.make_layer(
            {
                "class": "eval",
                "eval": "tf.clip_by_value(source(0), source(1), source(2))",
                "from": [x, clip_value_min, clip_value_max],
                "allow_broadcast_all_sources": allow_broadcast_all_sources,
            },
            name="clip_by_value",
        )

    @staticmethod
    def cumsum(source: Tensor, *, spatial_dim: Dim) -> Tensor:
        """cumsum"""
        return rfl.make_layer({"class": "cumsum", "from": source, "axis": spatial_dim}, name="cumsum")

    @staticmethod
    def matmul(a: Tensor, b: Tensor, *, reduce: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor:
        """matmul"""
        args = {}
        if not use_mask:
            args["use_mask"] = False
        return rfl.make_layer({"class": "dot", "from": [a, b], "reduce": reduce, **args}, name="matmul")

    @staticmethod
    def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor:
        """range over dim"""
        if not dtype and dim.dyn_size_ext:
            dtype = dim.dyn_size_ext.dtype
        if not dtype:
            dtype = rf.get_default_array_index_dtype()
        return rfl.make_layer(
            {
                "class": "range_in_axis",
                "from": _dims.get_dim_deps(dim),
                "axis": dim,
                "dtype": dtype,
                "sparse": dtype.startswith("int") or dtype.startswith("uint"),
            },
            name="range_over_dim",
        )

    @staticmethod
    def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Dim) -> Tensor:
        """
        :param source:
        :param in_dim:
        :param out_dim:
        :return: source with in_dim replaced by out_dim.
        """
        return rfl.make_layer(
            {"class": "reinterpret_data", "set_dim_tags": {in_dim: out_dim}, "from": source}, name="new_dim"
        )

    @staticmethod
    def set_sparse_dim(source: Tensor, sparse_dim: Dim) -> Tensor:
        """set sparse dim"""
        return rfl.make_layer(
            {"class": "reinterpret_data", "set_sparse": True, "set_sparse_dim": sparse_dim, "from": source},
            name="set_sparse_dim",
        )

    @staticmethod
    def reduce(source: Tensor, *, mode: str, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor:
        """Reduce"""
        assert mode in Backend._AllowedReduceModes
        kwargs = {}
        if not use_mask:
            kwargs["use_time_mask"] = False
        return rfl.make_layer(
            {"class": "reduce", "from": source, "mode": mode, "axis": axis, **kwargs}, name=f"reduce_{mode}"
        )

    # noinspection PyShadowingBuiltins
    @staticmethod
    def top_k(
        source: Tensor,
        *,
        axis: Union[Dim, Sequence[Dim]],
        k: Union[int, Tensor],
        k_dim: Optional[Dim] = None,
        sorted: bool = True,
    ) -> Tuple[Tensor, Union[Tensor, Sequence[Tensor]], Dim]:
        """top_k"""
        if not k_dim:
            k_dim = Dim(k, name="top-k-dim")
        values = rfl.make_layer(
            {
                "class": "top_k",
                "from": source,
                "axis": axis,
                "k": k,
                "k_dim": k_dim,
                "sorted": sorted,
            },
            name="top_k",
        )
        if isinstance(axis, (tuple, list)):
            axes = axis
            single_axis = False
        else:
            assert isinstance(axis, Dim)
            axes = [axis]
            single_axis = True
        indices = []
        for i, a in enumerate(axes):
            assert isinstance(a, Dim)
            sub_name = "indices" if single_axis else f"indices{i}"
            indices_data = values.copy_template(name=f"{values.name}_{sub_name}_{a.description}")
            indices_data.dtype = "int32"
            indices_data.sparse_dim = a
            # noinspection PyProtectedMember
            indices.append(rfl._get_sub_layer(values, sub_name, data=indices_data))
        if single_axis:
            indices = indices[0]
        return values, indices, k_dim

    @staticmethod
    @contextlib.contextmanager
    def random_journal_replay(journal: _random_journal.RandomJournal):
        """
        Replays the journal.
        At exit, the journal is cleared, and we check that we replayed everything.
        """
        try:
            ReturnnLayersBackend._random_journal = journal
            yield
        finally:
            ReturnnLayersBackend._random_journal = None

    _random_journal = None  # type: Optional[_random_journal.RandomJournal]

    @staticmethod
    def random(
        *,
        dims: Sequence[Dim],
        dtype: str,
        device: Optional[str] = None,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
        distribution: str,
        mean: Optional[Union[int, float, Tensor]] = None,
        stddev: Optional[Union[int, float, Tensor]] = None,
        bound: Optional[Union[int, float, Tensor]] = None,
        minval: Optional[Union[int, float, Tensor]] = None,
        maxval: Optional[Union[int, float, Tensor]] = None,
        seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
        algorithm: Optional[str] = None,
        explicit_state: Optional[Tensor] = None,
        auto_update_state: Optional[bool] = None,
        static: Optional[bool] = None,
        out: Optional[Tensor] = None,
    ) -> Tensor:
        """random"""
        if ReturnnLayersBackend._random_journal:
            recent = ReturnnLayersBackend._random_journal.get_recent_graph_reader_node_in_accessible_ctx()
            out = rfl.make_layer(
                {
                    "class": "eval",
                    "from": [recent] if recent else [],  # use as control dependency
                    "eval": _random_replay_eval,
                    "eval_locals": {"idx": ReturnnLayersBackend._random_journal.get_graph_reader_idx()},
                    "out_type": {"dims": dims, "dtype": dtype, "sparse_dim": sparse_dim, "feature_dim": feature_dim},
                },
                name="random_replay",
            )
            if out.control_flow_ctx and out.control_flow_ctx.is_loop():
                # Make sure it stays inside the loop.
                out.raw_tensor.layer_dict["from"].append(
                    rfl.PrevTensorRef.get_prev_ref(
                        cur_layer_name_ctx=out.raw_tensor,
                        initial=rf.zeros(dims, dtype=dtype, sparse_dim=sparse_dim, feature_dim=feature_dim),
                    )
                )
            ReturnnLayersBackend._random_journal.add_graph_reader_node(out)
            return out
        kwargs = {
            "mean": mean,
            "stddev": stddev,
            "bound": bound,
            "minval": minval,
            "maxval": maxval,
            "seed": seed,
            "algorithm": algorithm,
            "explicit_state": explicit_state,
            "auto_update_state": auto_update_state,
            "static": static,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return rfl.make_layer(
            {
                "class": "random",
                "shape": dims,
                "shape_deps": _dims.get_dim_deps(dims),
                "dtype": dtype,
                "sparse_dim": sparse_dim,
                "feature_dim": feature_dim,
                "distribution": distribution,
                "stop_grad": True,
                **kwargs,
            },
            name="random",
        )

    @staticmethod
    def masked_select(
        tensor: Tensor, *, mask: Tensor, dims: Sequence[Dim], out_dim: Optional[Dim] = None
    ) -> Tuple[Tensor, Dim]:
        """
        :param tensor:
        :param mask:
        :param dims: the order of the dims defines the format. those dims should be exactly the dims of the mask.
        :param out_dim:
        :return: tensor where all dims in mask/dims are removed and replaced by a new dim.
            the new dim is also returned.
            if mask==True for all elements, the returned tensor would be simply the flattened input tensor.
        """
        assert mask.dtype == "bool"
        assert set(mask.dims) == set(dims)
        assert set(mask.dims).issubset(set(tensor.dims))
        if not out_dim:
            out_dim = Dim(None, name="mask")
        return (
            rfl.make_layer(
                {"class": "boolean_mask", "from": tensor, "mask": mask, "dims": dims, "out_dim": out_dim},
                name="boolean_mask",
            ),
            out_dim,
        )

    @staticmethod
    def batch_norm(
        source: Tensor,
        *,
        in_dim: Union[Dim, Sequence[Dim]],
        running_mean: Tensor,
        running_variance: Tensor,
        gamma: Optional[Tensor],
        beta: Optional[Tensor],
        epsilon: float,
        momentum: float,
        affine: bool,
        use_mask: bool,
    ) -> Tensor:
        """batch norm"""
        reuse_params = {
            "batch_norm/v2_mean": running_mean,
            "batch_norm/v2_variance": running_variance,
        }
        if affine:
            reuse_params["batch_norm/v2_gamma"] = gamma
            reuse_params["batch_norm/v2_beta"] = beta
        reuse_params = {"map": {k: {"layer_output": v} for k, v in reuse_params.items()}}
        return rfl.make_layer(
            {
                "class": "batch_norm",
                "from": source,
                "in_dim": in_dim,
                "use_std": affine,
                "use_shift": affine,
                "param_version": 2,
                "reuse_params": reuse_params,
                "momentum": momentum,
                "epsilon": epsilon,
                "masked_time": use_mask,
            },
            name="batch_norm",
        )

    # noinspection PyShadowingBuiltins
    @staticmethod
    def conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],  # to have the order well-defined
        padding: str,
        strides: Optional[Union[int, Sequence[int]]] = None,
        dilation_rate: Optional[Union[int, Sequence[int]]] = None,
        groups: Optional[int] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """conv"""
        if not out_spatial_dims:
            out_spatial_dims = rf.make_conv_out_spatial_dims(
                description_prefix="conv",
                in_spatial_dims=in_spatial_dims,
                filter_size=[d.dimension for d in filter_size],
                strides=strides or 1,
                dilation_rate=dilation_rate or 1,
                padding=padding,
            )
        layer_dict = {
            "class": "conv",
            "from": source,
            "in_dim": in_dim,
            "in_spatial_dims": in_spatial_dims,
            "out_dim": out_dim,
            "out_spatial_dims": out_spatial_dims,
            "filter_size": filter_size,
            "padding": padding,
        }
        if strides:
            layer_dict["strides"] = strides
        if dilation_rate:
            layer_dict["dilation_rate"] = dilation_rate
        if groups:
            layer_dict["groups"] = groups
        layer_dict.update({"filter": filter, "with_bias": bias is not None})
        if bias is not None:
            layer_dict["bias"] = bias
        out = rfl.make_layer(layer_dict, name="conv")
        return out, out_spatial_dims

    # noinspection PyShadowingBuiltins
    @staticmethod
    def transposed_conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],
        padding: str,
        remove_padding: Union[Sequence[int], int] = 0,
        output_padding: Optional[Union[Sequence[Optional[int]], int]] = None,
        strides: Optional[Sequence[int]] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """transposed conv"""
        if not out_spatial_dims:
            out_spatial_dims = [Dim(None, name=f"out-spatial-dim{i}") for i, s in enumerate(filter_size)]
            for i in range(len(filter_size)):
                s = filter_size[i].dimension if not strides else strides[i]
                if filter_size[i].dimension == s == 1 or (s == 1 and padding.lower() == "same"):
                    out_spatial_dims[i] = in_spatial_dims[i]
        layer_dict = {
            "class": "transposed_conv",
            "from": source,
            "in_dim": in_dim,
            "in_spatial_dims": in_spatial_dims,
            "out_dim": out_dim,
            "out_spatial_dims": out_spatial_dims,
            "filter_size": filter_size,
            "padding": padding,
        }
        if remove_padding:
            layer_dict["remove_padding"] = remove_padding
        if output_padding:
            layer_dict["output_padding"] = output_padding
        if strides:
            layer_dict["strides"] = strides
        layer_dict.update({"filter": filter, "with_bias": bias is not None})
        if bias is not None:
            layer_dict["bias"] = bias
        out = rfl.make_layer(layer_dict, name="transposed_conv")
        return out, out_spatial_dims

    @staticmethod
    def pool(
        source: Tensor,
        *,
        mode: str,
        pool_size: Sequence[int],
        padding: str = "valid",
        dilation_rate: Union[Sequence[int], int] = 1,
        strides: Sequence[int],
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """pool"""
        if out_spatial_dims is None:
            out_spatial_dims = rf.make_conv_out_spatial_dims(
                description_prefix="pool",
                in_spatial_dims=in_spatial_dims,
                filter_size=pool_size,
                strides=strides,
                dilation_rate=dilation_rate,
                padding=padding,
            )
        other_dims = [d for d in source.dims if d not in in_spatial_dims and not d.is_batch_dim()]
        dummy_in_dim = None
        if source.feature_dim and source.feature_dim in other_dims:
            in_dim = source.feature_dim
        elif other_dims:
            in_dim = other_dims[-1]
        else:
            # PoolLayer currently needs some in_dim.
            dummy_in_dim = Dim(1, name="dummy_in")
            in_dim = dummy_in_dim
            source = rf.expand_dim(source, dim=dummy_in_dim)
        assert source.have_batch_axis(), "PoolLayer without batch dim not implemented"
        args = {
            "mode": mode,
            "pool_size": pool_size,
            "padding": padding,
            "dilation_rate": dilation_rate,
            "strides": strides,
            "in_spatial_dims": in_spatial_dims,
            "out_spatial_dims": out_spatial_dims,
            "in_dim": in_dim,  # it does not really matter, but we need sth currently
        }
        layer = rfl.make_layer({"class": "pool", "from": source, **args}, name="pool")
        if dummy_in_dim:
            layer = rf.squeeze(layer, axis=dummy_in_dim)
        if source.feature_dim != in_dim:
            # We want that the feature-dim stays consistent. PoolLayer currently just sets it to the in_dim.
            layer = rfl.make_layer(
                {"class": "reinterpret_data", "from": layer, "set_axes": {"F": source.feature_dim}},
                name="pool_reset_feature",
            )
        return layer, out_spatial_dims

    @staticmethod
    def stft(
        x: Tensor,
        *,
        in_spatial_dim: Dim,
        frame_step: int,
        frame_length: int,
        fft_length: int,
        window_use_frame_length: bool = True,
        align_window_left: bool = True,
        window_enforce_even: bool = True,
        out_spatial_dim: Dim,
        out_dim: Dim,
    ) -> Tensor:
        """stft"""
        if frame_length < fft_length:
            assert window_use_frame_length, "not implemented otherwise"
            assert align_window_left, "not implemented otherwise"
        if fft_length % 2 != 0:
            assert window_enforce_even, "not implemented otherwise"
        return rfl.make_layer(
            {
                "class": "stft",
                "from": x,
                "in_spatial_dims": [in_spatial_dim],
                "out_spatial_dims": [out_spatial_dim],
                "out_dim": out_dim,
                "frame_shift": frame_step,
                "frame_size": frame_length,
                "fft_size": fft_length,
            }
        )

    @staticmethod
    def lstm(
        source: Tensor,
        *,
        state_h: Tensor,
        state_c: Tensor,
        ff_weight: Tensor,
        rec_weight: Tensor,
        bias: Tensor,
        spatial_dim: Dim,
        in_dim: Dim,
        out_dim: Dim,
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        """
        :return: output, (h, c)
        """
        from ._utils import get_last_hidden_state

        # PyTorch (cuDNN) weights are in ifco order (?),
        # which we defined as the standard for the RF.
        # Also, they are (in_dim|out_dim, 4*out_dim).
        # RETURNN NativeLstm2 weight order: cell-in + input, forget and output gates (cifo).
        # And they are (4*out_dim, in_dim|out_dim).
        # So we need to reorder the params (ifco->cifo) and transpose them.
        # See also CustomCheckpointLoader and convert_cudnn_canonical_to_lstm_block.
        # TODO: ideally, we would create a new NativeLstm variant which just uses the same order.
        # Need new dim because after the split, we would get (out_dim,out_dim) which is ambiguous...
        out_dim_ = out_dim.copy(same_as_self=False, description="(out-dim)")
        rec_weight_ = rf.split(rec_weight, axis=4 * out_dim, out_dims=[out_dim_] * 4)
        ff_weight_ = rf.split(ff_weight, axis=4 * out_dim, out_dims=[out_dim_] * 4)
        bias_ = rf.split(bias, axis=4 * out_dim, out_dims=[out_dim_] * 4)
        rec_weight, _ = rf.concat(
            (rec_weight_[2], out_dim_),
            (rec_weight_[0], out_dim_),
            (rec_weight_[1], out_dim_),
            (rec_weight_[3], out_dim_),
        )
        ff_weight, _ = rf.concat(
            (ff_weight_[2], out_dim_), (ff_weight_[0], out_dim_), (ff_weight_[1], out_dim_), (ff_weight_[3], out_dim_)
        )
        bias, _ = rf.concat((bias_[2], out_dim_), (bias_[0], out_dim_), (bias_[1], out_dim_), (bias_[3], out_dim_))
        rec_weight, _ = rf.replace_dim(rec_weight, in_dim=4 * out_dim_, out_dim=4 * out_dim)
        ff_weight, _ = rf.replace_dim(ff_weight, in_dim=4 * out_dim_, out_dim=4 * out_dim)
        bias, _ = rf.replace_dim(bias, in_dim=4 * out_dim_, out_dim=4 * out_dim)

        output = rfl.make_layer(
            {
                "class": "rec",
                "from": source,
                "in_dim": in_dim,
                "axis": spatial_dim,
                "out_dim": out_dim,
                "unit": "lstm",
                "reuse_params": {
                    "map": {
                        "W_re": {"layer_output": rec_weight, "shape": (out_dim, 4 * out_dim)},
                        "W": {"layer_output": ff_weight, "shape": (in_dim, 4 * out_dim)},
                        "b": {"layer_output": bias, "shape": (4 * out_dim,)},
                    }
                },
                "initial_state": {"h": state_h, "c": state_c},
            },
            name="lstm",
        )
        h = get_last_hidden_state(output, out_dim=out_dim, key="h")
        c = get_last_hidden_state(output, out_dim=out_dim, key="c")
        return output, (h, c)


def _random_replay_eval(*, self, source, idx: int, **_kwargs):
    from returnn.tf.layers.basic import LayerBase

    assert isinstance(self, LayerBase)
    idx  # noqa  # unused - this can be helpful for debugging that the execution order is correct

    def _py_func() -> numpy.ndarray:
        # noinspection PyProtectedMember
        elem = ReturnnLayersBackend._random_journal.get_next(new_out_template=self.output)
        assert isinstance(elem.out, Tensor)
        assert isinstance(elem.out.raw_tensor, numpy.ndarray)
        return elem.out.raw_tensor

    def _func() -> tf.Tensor:
        (out,) = tf.numpy_function(_py_func, [], [self.output.dtype])
        assert isinstance(out, tf.Tensor)
        out.set_shape(self.output.batch_shape)
        return out

    with (
        tf.control_dependencies([source(i, auto_convert=False) for i in range(len(self.sources))])
        if self.sources
        else contextlib.nullcontext()
    ):
        return _func()



#!/usr/bin/env python3

"""
This script can import a Blocks MT model into Returnn.
It currently assumes a specific Returnn network topology with specific layer names.
Example Returnn network topology:

.. code-block:: python

    network = {
    "source_embed": {"class": "linear", "activation": None, "with_bias": False, "n_out": 620},

    "lstm0_fw" : { "class": "rec", "unit": "standardlstm", "unit_opts": {"use_peepholes": True, "forget_bias": 0.0}, "initial_state": "var", "n_out" : 1000, "direction": 1, "from": ["source_embed"] },
    "lstm0_bw" : { "class": "rec", "unit": "standardlstm", "unit_opts": {"use_peepholes": True, "forget_bias": 0.0}, "initial_state": "var", "n_out" : 1000, "direction": -1, "from": ["source_embed"] },

    "lstm1_fw" : { "class": "rec", "unit": "standardlstm", "unit_opts": {"use_peepholes": True, "forget_bias": 0.0}, "initial_state": "var", "n_out" : 1000, "direction": 1, "from": ["lstm0_fw", "lstm0_bw"] },
    "lstm1_bw" : { "class": "rec", "unit": "standardlstm", "unit_opts": {"use_peepholes": True, "forget_bias": 0.0}, "initial_state": "var", "n_out" : 1000, "direction": -1, "from": ["lstm0_fw", "lstm0_bw"] },

    "encoder": {"class": "copy", "from": ["lstm1_fw", "lstm1_bw"]},
    "enc_ctx": {"class": "linear", "activation": None, "with_bias": True, "from": ["encoder"], "n_out": 1000},  # preprocessed_attended in Blocks
    "fertility": {"class": "linear", "activation": "sigmoid", "with_bias": False, "from": ["encoder"], "n_out": 1},

    "output": {"class": "rec", "from": [], "unit": {
        'output': {'class': 'choice', 'target': 'classes', 'beam_size': beam_size, 'from': ["output_prob"], "initial_output": 0, "length_normalization": False},
        "end": {"class": "compare", "from": ["output"], "value": 0},
        'target_embed': {'class': 'linear', 'activation': None, "with_bias": False, 'from': ['output'], "n_out": 621, "initial_output": 0},  # feedback_input
        "weight_feedback": {"class": "linear", "activation": None, "with_bias": False, "from": ["prev:accum_att_weights"], "n_out": 1000},
        "prev_s_state": {"class": "get_last_hidden_state", "from": ["prev:s"], "n_out": 2000},
        "prev_s_transformed": {"class": "linear", "activation": None, "with_bias": False, "from": ["prev_s_state"], "n_out": 1000},
        "energy_in": {"class": "combine", "kind": "add", "from": ["base:enc_ctx", "weight_feedback", "prev_s_transformed"], "n_out": 1000},
        "energy_tanh": {"class": "activation", "activation": "tanh", "from": ["energy_in"]},
        "energy": {"class": "linear", "activation": None, "with_bias": False, "from": ["energy_tanh"], "n_out": 1},  # (B, enc-T, 1)
        "att_weights": {"class": "softmax_over_spatial", "from": ["energy"]},  # (B, enc-T, 1)
        "accum_att_weights": {"class": "eval", "from": ["prev:accum_att_weights", "att_weights", "base:fertility"],
            "eval": "source(0) + source(1) / (2.0 * source(2))", "out_type": {"dim": 1, "shape": (None, 1)}},
        "att": {"class": "generic_attention", "weights": "att_weights", "base": "base:encoder"},
        "s": {"class": "rnn_cell", "unit": "standardlstm", "unit_opts": {"use_peepholes": True, "forget_bias": 0.0},
            "initial_state": "var", "from": ["target_embed", "att"], "n_out": 1000},  # transform
        "readout_in": {"class": "linear", "from": ["prev:s", "prev:target_embed", "att"], "activation": None, "n_out": 1000},  # merge + post_merge bias
        "readout": {"class": "reduce_out", "mode": "max", "num_pieces": 2, "from": ["readout_in"]},
        "output_prob": {"class": "softmax", "from": ["readout"], "target": "classes", "loss": "ce"}
    }, "target": "classes", "max_seq_len": 75},

    "decision": {
        "class": "decide", "from": ["output"], "length_normalization": True,
        "loss": "edit_distance", "target": "classes",
        "loss_opts": {
            #"debug_print": True
            }
        }
    }

"""

from __future__ import annotations

import os
import sys
import numpy
import re
from pprint import pprint
from numpy.testing import assert_almost_equal
import tensorflow as tf
import pickle

my_dir = os.path.dirname(os.path.abspath(__file__))
returnn_dir = os.path.dirname(my_dir)
sys.path.insert(0, returnn_dir)

from returnn.util import better_exchook
import returnn.__main__ as rnn
import returnn.util.basic as util
from returnn.tf.network import TFNetwork
from returnn.tf.layers.basic import SourceLayer, LayerBase, LinearLayer
from returnn.tf.layers.rec import ChoiceLayer


def get_network():
    """
    :rtype: TFNetwork
    """
    return rnn.engine.network


def get_input_layers():
    """
    :rtype: list[LayerBase]
    """
    ls = []
    for layer in get_network().layers.values():
        if len(layer.sources) != 1:
            continue
        if isinstance(layer.sources[0], SourceLayer):
            ls.append(layer)
    return ls


def find_our_input_embed_layer():
    """
    :rtype: LinearLayer
    """
    input_layers = get_input_layers()
    assert len(input_layers) == 1
    layer = input_layers[0]
    assert isinstance(layer, LinearLayer)
    return layer


def get_in_hierarchy(name, hierarchy):
    """
    :param str name: e.g. "decoder/sequencegenerator"
    :param dict[str,dict[str]] hierarchy: nested hierarchy
    :rtype: dict[str,dict[str]]
    """
    if "/" in name:
        name, rest = name.split("/", 1)
    else:
        rest = None
    if rest is None:
        return hierarchy[name]
    else:
        return get_in_hierarchy(rest, hierarchy[name])


def main():
    rnn.init(
        command_line_options=sys.argv[1:],
        config_updates={
            "task": "nop",
            "log": None,
            "device": "cpu",
            "allow_random_model_init": True,
            "debug_add_check_numerics_on_output": False,
        },
        extra_greeting="Import Blocks MT model.",
    )
    assert util.BackendEngine.is_tensorflow_selected()
    config = rnn.config

    # Load Blocks MT model params.
    if not config.has("blocks_mt_model"):
        print("Please provide the option blocks_mt_model.")
        sys.exit(1)
    blocks_mt_model_fn = config.value("blocks_mt_model", "")
    assert blocks_mt_model_fn
    assert os.path.exists(blocks_mt_model_fn)
    if os.path.isdir(blocks_mt_model_fn):
        blocks_mt_model_fn += "/params.npz"
        assert os.path.exists(blocks_mt_model_fn)

    dry_run = config.bool("dry_run", False)
    if dry_run:
        our_model_fn = None
        print("Dry-run, will not save model.")
    else:
        our_model_fn = config.value("model", "returnn-model") + ".imported"
        print("Will save Returnn model as %s." % our_model_fn)
        assert os.path.exists(os.path.dirname(our_model_fn) or "."), "model-dir does not exist"
        assert not os.path.exists(our_model_fn + util.get_model_filename_postfix()), "model-file already exists"

    blocks_mt_model = numpy.load(blocks_mt_model_fn)
    assert isinstance(blocks_mt_model, numpy.lib.npyio.NpzFile), "did not expect type %r in file %r" % (
        type(blocks_mt_model),
        blocks_mt_model_fn,
    )
    print("Params found in Blocks model:")
    blocks_params = {}  # type: dict[str,numpy.ndarray]
    blocks_params_hierarchy = {}  # type: dict[str,dict[str]]
    blocks_total_num_params = 0
    for key in sorted(blocks_mt_model.keys()):
        value = blocks_mt_model[key]
        key = key.replace("-", "/")
        assert key[0] == "/"
        key = key[1:]
        blocks_params[key] = value
        print("  %s: %s, %s" % (key, value.shape, value.dtype))
        blocks_total_num_params += numpy.prod(value.shape)
        d = blocks_params_hierarchy
        for part in key.split("/"):
            d = d.setdefault(part, {})
    print("Blocks total num params: %i" % blocks_total_num_params)

    # Init our network structure.
    from returnn.tf.layers.rec import _SubnetworkRecCell

    _SubnetworkRecCell._debug_out = []  # enable for debugging intermediate values below
    ChoiceLayer._debug_out = []  # also for debug outputs of search
    rnn.engine.use_search_flag = True  # construct the net as in search
    rnn.engine.init_network_from_config()
    print("Our network model params:")
    our_params = {}  # type: dict[str,tf.Variable]
    our_total_num_params = 0
    for v in rnn.engine.network.get_params_list():
        key = v.name[:-2]
        our_params[key] = v
        print("  %s: %s, %s" % (key, v.shape, v.dtype.base_dtype.name))
        our_total_num_params += numpy.prod(v.shape.as_list())
    print("Our total num params: %i" % our_total_num_params)

    # Now matching...
    blocks_used_params = set()  # type: set[str]
    our_loaded_params = set()  # type: set[str]

    def import_var(our_var, blocks_param):
        """
        :param tf.Variable our_var:
        :param str|numpy.ndarray blocks_param:
        """
        assert isinstance(our_var, tf.Variable)
        if isinstance(blocks_param, str):
            blocks_param = load_blocks_var(blocks_param)
        assert isinstance(blocks_param, numpy.ndarray)
        assert tuple(our_var.shape.as_list()) == blocks_param.shape
        our_loaded_params.add(our_var.name[:-2])
        our_var.load(blocks_param, session=rnn.engine.tf_session)

    def load_blocks_var(blocks_param_name):
        """
        :param str blocks_param_name:
        :rtype: numpy.ndarray
        """
        assert isinstance(blocks_param_name, str)
        assert blocks_param_name in blocks_params
        blocks_used_params.add(blocks_param_name)
        return blocks_params[blocks_param_name]

    enc_name = "bidirectionalencoder"
    enc_embed_name = "EncoderLookUp0.W"
    assert enc_name in blocks_params_hierarchy
    assert enc_embed_name in blocks_params_hierarchy[enc_name]  # input embedding
    num_encoder_layers = max(
        [
            int(re.match(".*([0-9]+)", s).group(1))
            for s in blocks_params_hierarchy[enc_name]
            if s.startswith("EncoderBidirectionalLSTM")
        ]
    )
    blocks_input_dim, blocks_input_embed_dim = blocks_params["%s/%s" % (enc_name, enc_embed_name)].shape
    print("Blocks input dim: %i, embed dim: %i" % (blocks_input_dim, blocks_input_embed_dim))
    print("Blocks num encoder layers: %i" % num_encoder_layers)
    expected_enc_entries = ["EncoderLookUp0.W"] + [
        "EncoderBidirectionalLSTM%i" % i for i in range(1, num_encoder_layers + 1)
    ]
    assert set(expected_enc_entries) == set(blocks_params_hierarchy[enc_name].keys())

    our_input_layer = find_our_input_embed_layer()
    assert our_input_layer.input_data.dim == blocks_input_dim
    assert our_input_layer.output.dim == blocks_input_embed_dim
    assert not our_input_layer.with_bias
    import_var(our_input_layer.params["W"], "%s/%s" % (enc_name, enc_embed_name))

    dec_name = "decoder/sequencegenerator"
    dec_hierarchy_base = get_in_hierarchy(dec_name, blocks_params_hierarchy)
    assert set(dec_hierarchy_base.keys()) == {"att_trans", "readout"}
    dec_embed_name = "readout/lookupfeedbackwmt15/lookuptable.W"
    get_in_hierarchy(dec_embed_name, dec_hierarchy_base)  # check

    for i in range(num_encoder_layers):
        # Assume standard LSTMCell.
        # i = input_gate, j = new_input, f = forget_gate, o = output_gate
        # lstm_matrix = self._linear1([inputs, m_prev])
        # i, j, f, o = array_ops.split(value=lstm_matrix, num_or_size_splits=4, axis=1)
        # bias (4*in), kernel (in+out,4*out), w_(f|i|o)_diag (out)
        # prefix: rec/rnn/lstm_cell
        # Blocks: gate-in, gate-forget, next-in, gate-out
        for direction in ("fwd", "bwd"):
            our_layer = get_network().layers["lstm%i_%s" % (i, direction[:2])]
            blocks_prefix = "bidirectionalencoder/EncoderBidirectionalLSTM%i" % (i + 1,)
            # (in,out*4), (out*4,)
            W_in, b = [
                load_blocks_var(
                    "%s/%s_fork/fork_inputs.%s" % (blocks_prefix, {"bwd": "back", "fwd": "fwd"}[direction], p)
                )
                for p in ("W", "b")
            ]
            W_re = load_blocks_var(
                "%s/bidirectionalseparateparameters/%s.W_state"
                % (blocks_prefix, {"fwd": "forward", "bwd": "backward"}[direction])
            )
            W = numpy.concatenate([W_in, W_re], axis=0)
            b = lstm_vec_blocks_to_tf(b)
            W = lstm_vec_blocks_to_tf(W)
            import_var(our_layer.params["rnn/lstm_cell/bias"], b)
            import_var(our_layer.params["rnn/lstm_cell/kernel"], W)
            import_var(
                our_layer.params["initial_c"],
                "%s/bidirectionalseparateparameters/%s.initial_cells"
                % (blocks_prefix, {"fwd": "forward", "bwd": "backward"}[direction]),
            )
            import_var(
                our_layer.params["initial_h"],
                "%s/bidirectionalseparateparameters/%s.initial_state"
                % (blocks_prefix, {"fwd": "forward", "bwd": "backward"}[direction]),
            )
            for s1, s2 in [
                ("W_cell_to_in", "w_i_diag"),
                ("W_cell_to_forget", "w_f_diag"),
                ("W_cell_to_out", "w_o_diag"),
            ]:
                import_var(
                    our_layer.params["rnn/lstm_cell/%s" % s2],
                    "%s/bidirectionalseparateparameters/%s.%s"
                    % (blocks_prefix, {"fwd": "forward", "bwd": "backward"}[direction], s1),
                )
    import_var(
        get_network().layers["enc_ctx"].params["W"],
        "decoder/sequencegenerator/att_trans/attention/encoder_state_transformer.W",
    )
    import_var(
        get_network().layers["enc_ctx"].params["b"],
        "decoder/sequencegenerator/att_trans/attention/encoder_state_transformer.b",
    )
    import_var(our_params["output/rec/s/initial_c"], "decoder/sequencegenerator/att_trans/lstm_decoder.initial_cells")
    import_var(our_params["output/rec/s/initial_h"], "decoder/sequencegenerator/att_trans/lstm_decoder.initial_state")
    import_var(
        our_params["output/rec/weight_feedback/W"],
        "decoder/sequencegenerator/att_trans/attention/sum_alignment_transformer.W",
    )
    import_var(
        our_params["output/rec/target_embed/W"], "decoder/sequencegenerator/readout/lookupfeedbackwmt15/lookuptable.W"
    )
    import_var(our_params["fertility/W"], "decoder/sequencegenerator/att_trans/attention/fertility_transformer.W")
    import_var(our_params["output/rec/energy/W"], "decoder/sequencegenerator/att_trans/attention/energy_comp/linear.W")
    prev_s_trans_W_states = load_blocks_var(
        "decoder/sequencegenerator/att_trans/attention/state_trans/transform_states.W"
    )
    prev_s_trans_W_cells = load_blocks_var(
        "decoder/sequencegenerator/att_trans/attention/state_trans/transform_cells.W"
    )
    prev_s_trans_W = numpy.concatenate([prev_s_trans_W_cells, prev_s_trans_W_states], axis=0)
    import_var(our_params["output/rec/prev_s_transformed/W"], prev_s_trans_W)
    import_var(
        our_params["output/rec/s/rec/lstm_cell/bias"], numpy.zeros(our_params["output/rec/s/rec/lstm_cell/bias"].shape)
    )
    dec_lstm_kernel_in_feedback = load_blocks_var(
        "decoder/sequencegenerator/att_trans/feedback_to_decoder/fork_inputs.W"
    )
    dec_lstm_kernel_in_ctx = load_blocks_var("decoder/sequencegenerator/att_trans/context_to_decoder/fork_inputs.W")
    dec_lstm_kernel_re = load_blocks_var("decoder/sequencegenerator/att_trans/lstm_decoder.W_state")
    dec_lstm_kernel = numpy.concatenate(
        [dec_lstm_kernel_in_feedback, dec_lstm_kernel_in_ctx, dec_lstm_kernel_re], axis=0
    )
    dec_lstm_kernel = lstm_vec_blocks_to_tf(dec_lstm_kernel)
    import_var(our_params["output/rec/s/rec/lstm_cell/kernel"], dec_lstm_kernel)
    for s1, s2 in [("W_cell_to_in", "w_i_diag"), ("W_cell_to_forget", "w_f_diag"), ("W_cell_to_out", "w_o_diag")]:
        import_var(
            our_params["output/rec/s/rec/lstm_cell/%s" % s2], "decoder/sequencegenerator/att_trans/lstm_decoder.%s" % s1
        )
    readout_in_W_states = load_blocks_var("decoder/sequencegenerator/readout/merge/transform_states.W")
    readout_in_W_feedback = load_blocks_var("decoder/sequencegenerator/readout/merge/transform_feedback.W")
    readout_in_W_att = load_blocks_var("decoder/sequencegenerator/readout/merge/transform_weighted_averages.W")
    readout_in_W = numpy.concatenate([readout_in_W_states, readout_in_W_feedback, readout_in_W_att], axis=0)
    import_var(our_params["output/rec/readout_in/W"], readout_in_W)
    import_var(
        our_params["output/rec/readout_in/b"],
        "decoder/sequencegenerator/readout/initializablefeedforwardsequence/maxout_bias.b",
    )
    import_var(
        our_params["output/rec/output_prob/W"],
        "decoder/sequencegenerator/readout/initializablefeedforwardsequence/softmax1.W",
    )
    import_var(
        our_params["output/rec/output_prob/b"],
        "decoder/sequencegenerator/readout/initializablefeedforwardsequence/softmax1.b",
    )

    print("Not initialized own params:")
    count = 0
    for key, v in sorted(our_params.items()):
        if key in our_loaded_params:
            continue
        print("  %s: %s, %s" % (key, v.shape, v.dtype.base_dtype.name))
        count += 1
    if not count:
        print("  None.")
    print("Not used Blocks params:")
    count = 0
    for key, value in sorted(blocks_params.items()):
        if key in blocks_used_params:
            continue
        print("  %s: %s, %s" % (key, value.shape, value.dtype))
        count += 1
    if not count:
        print("  None.")
    print("Done.")

    blocks_debug_dump_output = config.value("blocks_debug_dump_output", None)
    if blocks_debug_dump_output:
        print("Will read Blocks debug dump output from %r and compare with Returnn outputs." % blocks_debug_dump_output)
        blocks_initial_outputs = numpy.load("%s/initial_states_data.0.npz" % blocks_debug_dump_output)
        blocks_search_log = pickle.load(open("%s/search.log.pkl" % blocks_debug_dump_output, "rb"), encoding="bytes")
        blocks_search_log = {d[b"step"]: d for d in blocks_search_log}
        input_seq = blocks_initial_outputs["input"]
        beam_size, seq_len = input_seq.shape
        input_seq = input_seq[0]  # all the same, select beam 0
        assert isinstance(input_seq, numpy.ndarray)
        print("Debug input seq: %s" % input_seq.tolist())
        from returnn.datasets.generating import StaticDataset

        dataset = StaticDataset(
            data=[{"data": input_seq}],
            output_dim={"data": get_network().extern_data.get_default_input_data().get_kwargs()},
        )
        dataset.init_seq_order(epoch=0)
        extract_output_dict = {
            "enc_src_emb": get_network().layers["source_embed"].output.get_placeholder_as_batch_major(),
            "encoder": get_network().layers["encoder"].output.get_placeholder_as_batch_major(),
            "enc_ctx": get_network().layers["enc_ctx"].output.get_placeholder_as_batch_major(),
            "output": get_network().layers["output"].output.get_placeholder_as_batch_major(),
        }
        from returnn.tf.layers.basic import concat_sources

        for i in range(num_encoder_layers):
            extract_output_dict["enc_layer_%i" % i] = concat_sources(
                [get_network().layers["lstm%i_fw" % i], get_network().layers["lstm%i_bw" % i]]
            ).get_placeholder_as_batch_major()
        extract_output_dict["enc_layer_0_fwd"] = (
            get_network().layers["lstm0_fw"].output.get_placeholder_as_batch_major()
        )
        our_output = rnn.engine.run_single(dataset=dataset, seq_idx=0, output_dict=extract_output_dict)
        blocks_out = blocks_initial_outputs["bidirectionalencoder_EncoderLookUp0__EncoderLookUp0_apply_output"]
        our_out = our_output["enc_src_emb"]
        print("our enc emb shape:", our_out.shape)
        print("Blocks enc emb shape:", blocks_out.shape)
        assert our_out.shape[:2] == (1, seq_len)
        assert blocks_out.shape[:2] == (seq_len, beam_size)
        assert our_out.shape[2] == blocks_out.shape[2]
        assert_almost_equal(our_out[0], blocks_out[:, 0], decimal=5)
        blocks_lstm0_out_ref = calc_lstm(blocks_out[:, 0], blocks_params)
        blocks_lstm0_out = blocks_initial_outputs[
            "bidirectionalencoder_EncoderBidirectionalLSTM1_bidirectionalseparateparameters_forward__forward_apply_states"
        ]
        our_lstm0_out = our_output["enc_layer_0_fwd"]
        assert blocks_lstm0_out.shape == (seq_len, beam_size) + blocks_lstm0_out_ref.shape
        assert our_lstm0_out.shape == (1, seq_len) + blocks_lstm0_out_ref.shape
        assert_almost_equal(blocks_lstm0_out[0, 0], blocks_lstm0_out_ref, decimal=6)
        print("Blocks LSTM0 frame 0 matched to ref calc.")
        assert_almost_equal(our_lstm0_out[0, 0], blocks_lstm0_out_ref, decimal=6)
        print("Our LSTM0 frame 0 matched to ref calc.")
        for i in range(num_encoder_layers):
            blocks_out = blocks_initial_outputs[
                "bidirectionalencoder_EncoderBidirectionalLSTM%i_bidirectionalseparateparameters__bidirectionalseparateparameters_apply_output_0"
                % (i + 1,)
            ]
            our_out = our_output["enc_layer_%i" % i]
            print("our enc layer %i shape:" % i, our_out.shape)
            print("Blocks enc layer %i shape:" % i, blocks_out.shape)
            assert our_out.shape[:2] == (1, seq_len)
            assert blocks_out.shape[:2] == (seq_len, beam_size)
            assert our_out.shape[2] == blocks_out.shape[2]
            assert_almost_equal(our_out[0], blocks_out[:, 0], decimal=6)
        print("our encoder shape:", our_output["encoder"].shape)
        blocks_encoder_out = blocks_initial_outputs["bidirectionalencoder__bidirectionalencoder_apply_representation"]
        print("Blocks encoder shape:", blocks_encoder_out.shape)
        assert our_output["encoder"].shape[:2] == (1, seq_len)
        assert blocks_encoder_out.shape[:2] == (seq_len, beam_size)
        assert our_output["encoder"].shape[2] == blocks_encoder_out.shape[2]
        assert_almost_equal(our_output["encoder"][0], blocks_encoder_out[:, 0], decimal=6)
        blocks_first_frame_outputs = numpy.load("%s/next_states.0.npz" % blocks_debug_dump_output)
        blocks_enc_ctx_out = blocks_first_frame_outputs[
            "decoder_sequencegenerator_att_trans_attention__attention_preprocess_preprocessed_attended"
        ]
        our_enc_ctx_out = our_output["enc_ctx"]
        print("Blocks enc ctx shape:", blocks_enc_ctx_out.shape)
        assert blocks_enc_ctx_out.shape[:2] == (seq_len, beam_size)
        assert our_enc_ctx_out.shape[:2] == (1, seq_len)
        assert blocks_enc_ctx_out.shape[2:] == our_enc_ctx_out.shape[2:]
        assert_almost_equal(blocks_enc_ctx_out[:, 0], our_enc_ctx_out[0], decimal=5)
        fertility = numpy.dot(
            blocks_encoder_out[:, 0],
            blocks_params["decoder/sequencegenerator/att_trans/attention/fertility_transformer.W"],
        )
        fertility = sigmoid(fertility)
        assert fertility.shape == (seq_len, 1)
        fertility = fertility[:, 0]
        assert fertility.shape == (seq_len,)
        our_dec_outputs = {v["step"]: v for v in _SubnetworkRecCell._debug_out}
        assert our_dec_outputs
        print("our dec frame keys:", sorted(our_dec_outputs[0].keys()))
        our_dec_search_outputs = {v["step"]: v for v in ChoiceLayer._debug_out}
        assert our_dec_search_outputs
        print("our dec search frame keys:", sorted(our_dec_search_outputs[0].keys()))
        print("Blocks search frame keys:", sorted(blocks_search_log[0].keys()))
        dec_lookup = blocks_params["decoder/sequencegenerator/readout/lookupfeedbackwmt15/lookuptable.W"]
        last_lstm_state = blocks_params["decoder/sequencegenerator/att_trans/lstm_decoder.initial_state"]
        last_lstm_cells = blocks_params["decoder/sequencegenerator/att_trans/lstm_decoder.initial_cells"]
        last_accumulated_weights = numpy.zeros((seq_len,), dtype="float32")
        last_output = 0
        dec_seq_len = 0
        for dec_step in range(100):
            blocks_frame_state_outputs_fn = "%s/next_states.%i.npz" % (blocks_debug_dump_output, dec_step)
            blocks_frame_probs_outputs_fn = "%s/logprobs.%i.npz" % (blocks_debug_dump_output, dec_step)
            if dec_step > 3:
                if not os.path.exists(blocks_frame_state_outputs_fn) or not os.path.exists(
                    blocks_frame_probs_outputs_fn
                ):
                    print("Seq not ended yet but frame not found for step %i." % dec_step)
                    break
            blocks_frame_state_outputs = numpy.load(blocks_frame_state_outputs_fn)
            blocks_frame_probs_outputs = numpy.load(blocks_frame_probs_outputs_fn)
            blocks_search_frame = blocks_search_log[dec_step]
            our_dec_frame_outputs = our_dec_outputs[dec_step]
            assert our_dec_frame_outputs["step"] == dec_step
            assert our_dec_frame_outputs[":i.output"].tolist() == [dec_step]
            our_dec_search_frame_outputs = our_dec_search_outputs[dec_step]

            blocks_last_lstm_state = blocks_frame_probs_outputs[
                "decoder_sequencegenerator__sequencegenerator_generate_states"
            ]
            blocks_last_lstm_cells = blocks_frame_probs_outputs[
                "decoder_sequencegenerator__sequencegenerator_generate_cells"
            ]
            assert blocks_last_lstm_state.shape == (beam_size, last_lstm_state.shape[0])
            assert_almost_equal(blocks_last_lstm_state[0], last_lstm_state, decimal=5)
            assert_almost_equal(blocks_last_lstm_cells[0], last_lstm_cells, decimal=5)
            our_last_lstm_cells = our_dec_frame_outputs["prev:s.extra.state"][0]
            our_last_lstm_state = our_dec_frame_outputs["prev:s.extra.state"][1]
            assert our_last_lstm_state.shape == our_last_lstm_cells.shape == (beam_size, last_lstm_state.shape[0])
            assert_almost_equal(our_last_lstm_state[0], last_lstm_state, decimal=5)
            assert_almost_equal(our_last_lstm_cells[0], last_lstm_cells, decimal=5)
            our_last_s = our_dec_frame_outputs["prev:s.output"]
            assert our_last_s.shape == (beam_size, last_lstm_state.shape[0])
            assert_almost_equal(our_last_s[0], last_lstm_state, decimal=5)

            blocks_last_accum_weights = blocks_frame_probs_outputs[
                "decoder_sequencegenerator__sequencegenerator_generate_accumulated_weights"
            ]
            assert blocks_last_accum_weights.shape == (beam_size, seq_len)
            assert_almost_equal(blocks_last_accum_weights[0], last_accumulated_weights, decimal=5)
            our_last_accum_weights = our_dec_frame_outputs["prev:accum_att_weights.output"]
            assert our_last_accum_weights.shape == (beam_size, seq_len if dec_step > 0 else 1, 1)
            if dec_step > 0:
                assert_almost_equal(our_last_accum_weights[0, :, 0], last_accumulated_weights, decimal=4)
            else:
                assert_almost_equal(our_last_accum_weights[0, 0, 0], last_accumulated_weights.sum(), decimal=4)

            energy_sum = numpy.copy(blocks_enc_ctx_out[:, 0])  # (T,enc-ctx-dim)
            weight_feedback = numpy.dot(
                last_accumulated_weights[:, None],
                blocks_params["decoder/sequencegenerator/att_trans/attention/sum_alignment_transformer.W"],
            )
            energy_sum += weight_feedback
            transformed_states = numpy.dot(
                last_lstm_state[None, :],
                blocks_params["decoder/sequencegenerator/att_trans/attention/state_trans/transform_states.W"],
            )
            transformed_cells = numpy.dot(
                last_lstm_cells[None, :],
                blocks_params["decoder/sequencegenerator/att_trans/attention/state_trans/transform_cells.W"],
            )
            energy_sum += transformed_states + transformed_cells
            assert energy_sum.shape == (seq_len, blocks_enc_ctx_out.shape[-1])
            blocks_energy_sum_tanh = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_att_trans_attention_energy_comp_tanh__tanh_apply_output"
            ]
            assert blocks_energy_sum_tanh.shape == (seq_len, beam_size, energy_sum.shape[-1])
            assert_almost_equal(blocks_energy_sum_tanh[:, 0], numpy.tanh(energy_sum), decimal=5)
            assert our_dec_frame_outputs["weight_feedback.output"].shape == (
                beam_size,
                seq_len if dec_step > 0 else 1,
                blocks_enc_ctx_out.shape[-1],
            )
            assert our_dec_frame_outputs["prev_s_transformed.output"].shape == (beam_size, blocks_enc_ctx_out.shape[-1])
            our_energy_sum = our_dec_frame_outputs["energy_in.output"]
            assert our_energy_sum.shape == (beam_size, seq_len, blocks_enc_ctx_out.shape[-1])
            assert_almost_equal(our_energy_sum[0], energy_sum, decimal=4)
            blocks_energy = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_att_trans_attention_energy_comp__energy_comp_apply_output"
            ]
            assert blocks_energy.shape == (seq_len, beam_size, 1)
            energy = numpy.dot(
                numpy.tanh(energy_sum),
                blocks_params["decoder/sequencegenerator/att_trans/attention/energy_comp/linear.W"],
            )
            assert energy.shape == (seq_len, 1)
            assert_almost_equal(blocks_energy[:, 0], energy, decimal=4)
            our_energy = our_dec_frame_outputs["energy.output"]
            assert our_energy.shape == (beam_size, seq_len, 1)
            assert_almost_equal(our_energy[0], energy, decimal=4)
            weights = softmax(energy[:, 0])
            assert weights.shape == (seq_len,)
            our_weights = our_dec_frame_outputs["att_weights.output"]
            assert our_weights.shape == (beam_size, seq_len, 1)
            assert_almost_equal(our_weights[0, :, 0], weights, decimal=4)
            accumulated_weights = last_accumulated_weights + weights / (2.0 * fertility)
            assert accumulated_weights.shape == (seq_len,)
            # blocks_accumulated_weights = blocks_frame_probs_outputs["decoder_sequencegenerator_att_trans_attention__attention_take_glimpses_accumulated_weights"]
            # assert blocks_accumulated_weights.shape == (beam_size, seq_len)
            # assert_almost_equal(blocks_accumulated_weights[0], accumulated_weights, decimal=5)
            blocks_weights = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_att_trans_attention__attention_compute_weights_output_0"
            ]
            assert blocks_weights.shape == (seq_len, beam_size)
            assert_almost_equal(weights, blocks_weights[:, 0], decimal=4)
            our_accum_weights = our_dec_frame_outputs["accum_att_weights.output"]
            assert our_accum_weights.shape == (beam_size, seq_len, 1)
            weighted_avg = (weights[:, None] * blocks_encoder_out[:, 0]).sum(axis=0)  # att in our
            assert weighted_avg.shape == (blocks_encoder_out.shape[-1],)
            blocks_weighted_avg = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_att_trans_attention__attention_compute_weighted_averages_output_0"
            ]
            assert blocks_weighted_avg.shape == (beam_size, blocks_encoder_out.shape[-1])
            assert_almost_equal(blocks_weighted_avg[0], weighted_avg, decimal=4)
            our_att = our_dec_frame_outputs["att.output"]
            assert our_att.shape == (beam_size, blocks_encoder_out.shape[-1])
            assert_almost_equal(our_att[0], weighted_avg, decimal=4)

            blocks_last_output = blocks_frame_probs_outputs[
                "decoder_sequencegenerator__sequencegenerator_generate_outputs"
            ]
            assert blocks_last_output.shape == (beam_size,)
            assert max(blocks_last_output[0], 0) == last_output
            last_target_embed = dec_lookup[last_output]
            if dec_step == 0:
                last_target_embed = numpy.zeros_like(last_target_embed)
            our_last_target_embed = our_dec_frame_outputs["prev:target_embed.output"]
            assert our_last_target_embed.shape == (beam_size, dec_lookup.shape[-1])
            assert_almost_equal(our_last_target_embed[0], last_target_embed, decimal=4)

            readout_in_state = numpy.dot(
                last_lstm_state, blocks_params["decoder/sequencegenerator/readout/merge/transform_states.W"]
            )
            blocks_trans_state = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_readout_merge__merge_apply_states"
            ]
            assert blocks_trans_state.shape == (beam_size, last_lstm_state.shape[0])
            assert_almost_equal(blocks_trans_state[0], readout_in_state, decimal=4)
            readout_in_feedback = numpy.dot(
                last_target_embed, blocks_params["decoder/sequencegenerator/readout/merge/transform_feedback.W"]
            )
            blocks_trans_feedback = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_readout_merge__merge_apply_feedback"
            ]
            assert blocks_trans_feedback.shape == (beam_size, readout_in_feedback.shape[0])
            assert_almost_equal(blocks_trans_feedback[0], readout_in_feedback, decimal=4)
            readout_in_weighted_avg = numpy.dot(
                weighted_avg, blocks_params["decoder/sequencegenerator/readout/merge/transform_weighted_averages.W"]
            )
            blocks_trans_weighted_avg = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_readout_merge__merge_apply_weighted_averages"
            ]
            assert blocks_trans_weighted_avg.shape == (beam_size, readout_in_weighted_avg.shape[0])
            assert_almost_equal(blocks_trans_weighted_avg[0], readout_in_weighted_avg, decimal=4)
            readout_in = readout_in_state + readout_in_feedback + readout_in_weighted_avg
            blocks_readout_in = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_readout_merge__merge_apply_output"
            ]
            assert blocks_readout_in.shape == (beam_size, readout_in.shape[0])
            assert_almost_equal(blocks_readout_in[0], readout_in, decimal=4)
            readout_in += blocks_params[
                "decoder/sequencegenerator/readout/initializablefeedforwardsequence/maxout_bias.b"
            ]
            assert readout_in.shape == (
                blocks_params["decoder/sequencegenerator/readout/initializablefeedforwardsequence/maxout_bias.b"].shape[
                    0
                ],
            )
            our_readout_in = our_dec_frame_outputs["readout_in.output"]
            assert our_readout_in.shape == (beam_size, readout_in.shape[0])
            assert_almost_equal(our_readout_in[0], readout_in, decimal=4)
            readout = readout_in.reshape((readout_in.shape[0] // 2, 2)).max(axis=1)
            our_readout = our_dec_frame_outputs["readout.output"]
            assert our_readout.shape == (beam_size, readout.shape[0])
            assert_almost_equal(our_readout[0], readout, decimal=4)
            prob_logits = (
                numpy.dot(
                    readout,
                    blocks_params["decoder/sequencegenerator/readout/initializablefeedforwardsequence/softmax1.W"],
                )
                + blocks_params["decoder/sequencegenerator/readout/initializablefeedforwardsequence/softmax1.b"]
            )
            assert prob_logits.ndim == 1
            blocks_prob_logits = blocks_frame_probs_outputs[
                "decoder_sequencegenerator_readout__readout_readout_output_0"
            ]
            assert blocks_prob_logits.shape == (beam_size, prob_logits.shape[0])
            assert_almost_equal(blocks_prob_logits[0], prob_logits, decimal=4)
            output_prob = softmax(prob_logits)
            log_output_prob = log_softmax(prob_logits)
            assert_almost_equal(numpy.log(output_prob), log_output_prob, decimal=4)
            our_output_prob = our_dec_frame_outputs["output_prob.output"]
            assert our_output_prob.shape == (beam_size, output_prob.shape[0])
            assert_almost_equal(our_output_prob[0], output_prob, decimal=4)
            blocks_nlog_prob = blocks_frame_probs_outputs["logprobs"]
            assert blocks_nlog_prob.shape == (beam_size, output_prob.shape[0])
            assert_almost_equal(blocks_nlog_prob[0], -log_output_prob, decimal=4)
            assert_almost_equal(our_dec_search_frame_outputs["scores_in_orig"][0], output_prob, decimal=4)
            assert_almost_equal(blocks_search_frame[b"logprobs"][0], -log_output_prob, decimal=4)
            # for b in range(beam_size):
            #  assert_almost_equal(-numpy.log(our_output_prob[b]), blocks_frame_probs_outputs["logprobs"][b], decimal=4)
            ref_output = numpy.argmax(output_prob)
            # Note: Don't take the readout.emit outputs. They are randomly sampled.
            blocks_dec_output = blocks_search_frame[b"outputs"]
            assert blocks_dec_output.shape == (beam_size,)
            our_dec_output = our_dec_frame_outputs["output.output"]
            assert our_dec_output.shape == (beam_size,)
            print("Frame %i: Ref best greedy output symbol: %i" % (dec_step, int(ref_output)))
            print("Blocks labels:", blocks_dec_output.tolist())
            print("Our labels:", our_dec_output.tolist())
            # Well, the following two could be not true if all the other beams have much better scores,
            # but this is unlikely.
            assert ref_output in blocks_dec_output
            assert ref_output in our_dec_output
            if dec_step == 0:
                # This assumes that the results are ordered by score which might not be true (see tf.nn.top_k).
                assert blocks_dec_output[0] == our_dec_output[0] == ref_output
            # We assume that the best is the same. Note that this also might not be true if there are two equally best scores.
            # It also assumes that it's ordered by the score which also might not be true (see tf.nn.top_k).
            # For the same reason, the remaining list and entries might also not perfectly match.
            assert our_dec_output[0] == blocks_dec_output[0]
            # Just follow the first beam.
            ref_output = blocks_dec_output[0]
            assert our_dec_search_frame_outputs["src_beam_idxs"].shape == (1, beam_size)
            assert our_dec_search_frame_outputs["scores"].shape == (1, beam_size)
            print("Blocks src_beam_idxs:", blocks_search_frame[b"indexes"].tolist())
            print("Our src_beam_idxs:", our_dec_search_frame_outputs["src_beam_idxs"][0].tolist())
            print("Blocks scores:", blocks_search_frame[b"chosen_costs"].tolist())
            print("Our scores:", our_dec_search_frame_outputs["scores"][0].tolist())
            if list(our_dec_search_frame_outputs["src_beam_idxs"][0]) != list(blocks_search_frame[b"indexes"]):
                print("Warning, beams do not match.")
                print("Blocks scores base:", blocks_search_frame[b"scores_base"].flatten().tolist())
                print("Our scores base:", our_dec_search_frame_outputs["scores_base"].flatten().tolist())
                # print("Blocks score in orig top k:", sorted(blocks_search_frame[b'logprobs'].flatten())[:beam_size])
                # print("Our score in orig top k:", sorted(-numpy.log(our_dec_search_frame_outputs["scores_in_orig"].flatten()))[:beam_size])
                print(
                    "Blocks score in top k:",
                    sorted(
                        (blocks_search_frame[b"logprobs"] * blocks_search_log[dec_step - 1][b"mask"][:, None]).flatten()
                    )[:beam_size],
                )
                print("Our score in top k:", sorted(-our_dec_search_frame_outputs["scores_in"].flatten())[:beam_size])
                blocks_scores_combined = blocks_search_frame[b"next_costs"]
                our_scores_combined = our_dec_search_frame_outputs["scores_combined"]
                print("Blocks scores combined top k:", sorted(blocks_scores_combined.flatten())[:beam_size])
                print("Our neg scores combined top k:", sorted(-our_scores_combined.flatten())[:beam_size])
                # raise Exception("beams mismatch")
            assert our_dec_search_frame_outputs["src_beam_idxs"][0][0] == blocks_search_frame[b"indexes"][0]
            beam_idx = our_dec_search_frame_outputs["src_beam_idxs"][0][0]
            if beam_idx != 0:
                print("Selecting different beam: %i." % beam_idx)
                # Just overwrite the needed states by Blocks outputs.
                accumulated_weights = blocks_frame_state_outputs[
                    "decoder_sequencegenerator_att_trans_attention__attention_take_glimpses_accumulated_weights"
                ][0]
                weighted_avg = blocks_frame_state_outputs[
                    "decoder_sequencegenerator__sequencegenerator_generate_weighted_averages"
                ][0]
                last_lstm_state = blocks_frame_state_outputs[
                    "decoder_sequencegenerator__sequencegenerator_generate_states"
                ][0]
                last_lstm_cells = blocks_frame_state_outputs[
                    "decoder_sequencegenerator__sequencegenerator_generate_cells"
                ][0]

            # From now on, use blocks_frame_state_outputs instead of blocks_frame_probs_outputs because
            # it will have the beam reordered.
            blocks_target_emb = blocks_frame_state_outputs[
                "decoder_sequencegenerator_fork__fork_apply_feedback_decoder_input"
            ]
            assert blocks_target_emb.shape == (beam_size, dec_lookup.shape[1])
            target_embed = dec_lookup[ref_output]
            assert target_embed.shape == (dec_lookup.shape[1],)
            assert_almost_equal(blocks_target_emb[0], target_embed)

            feedback_to_decoder = numpy.dot(
                target_embed, blocks_params["decoder/sequencegenerator/att_trans/feedback_to_decoder/fork_inputs.W"]
            )
            context_to_decoder = numpy.dot(
                weighted_avg, blocks_params["decoder/sequencegenerator/att_trans/context_to_decoder/fork_inputs.W"]
            )
            lstm_z = feedback_to_decoder + context_to_decoder
            assert (
                lstm_z.shape
                == feedback_to_decoder.shape
                == context_to_decoder.shape
                == (last_lstm_state.shape[-1] * 4,)
            )
            blocks_feedback_to_decoder = blocks_frame_state_outputs[
                "decoder_sequencegenerator_att_trans_feedback_to_decoder__feedback_to_decoder_apply_inputs"
            ]
            blocks_context_to_decoder = blocks_frame_state_outputs[
                "decoder_sequencegenerator_att_trans_context_to_decoder__context_to_decoder_apply_inputs"
            ]
            assert (
                blocks_feedback_to_decoder.shape
                == blocks_context_to_decoder.shape
                == (beam_size, last_lstm_state.shape[-1] * 4)
            )
            assert_almost_equal(blocks_feedback_to_decoder[0], feedback_to_decoder, decimal=4)
            assert_almost_equal(blocks_context_to_decoder[0], context_to_decoder, decimal=4)
            lstm_state, lstm_cells = calc_raw_lstm(
                lstm_z,
                blocks_params=blocks_params,
                prefix="decoder/sequencegenerator/att_trans/lstm_decoder.",
                last_state=last_lstm_state,
                last_cell=last_lstm_cells,
            )
            assert lstm_state.shape == last_lstm_state.shape == lstm_cells.shape == last_lstm_cells.shape
            blocks_lstm_state = blocks_frame_state_outputs[
                "decoder_sequencegenerator_att_trans_lstm_decoder__lstm_decoder_apply_states"
            ]
            blocks_lstm_cells = blocks_frame_state_outputs[
                "decoder_sequencegenerator_att_trans_lstm_decoder__lstm_decoder_apply_cells"
            ]
            assert blocks_lstm_state.shape == blocks_lstm_cells.shape == (beam_size, last_lstm_state.shape[-1])
            assert_almost_equal(blocks_lstm_state[0], lstm_state, decimal=4)
            assert_almost_equal(blocks_lstm_cells[0], lstm_cells, decimal=4)
            our_lstm_cells = our_dec_frame_outputs["s.extra.state"][0]
            our_lstm_state = our_dec_frame_outputs["s.extra.state"][1]
            assert our_lstm_state.shape == our_lstm_cells.shape == (beam_size, lstm_state.shape[0])
            assert_almost_equal(our_lstm_state[0], lstm_state, decimal=4)
            assert_almost_equal(our_lstm_cells[0], lstm_cells, decimal=4)
            our_s = our_dec_frame_outputs["s.output"]
            assert our_s.shape == (beam_size, lstm_state.shape[0])
            assert_almost_equal(our_s[0], lstm_state, decimal=4)

            last_accumulated_weights = accumulated_weights
            last_lstm_state = lstm_state
            last_lstm_cells = lstm_cells
            last_output = ref_output
            if last_output == 0:
                print("Sequence finished, seq len %i." % dec_step)
                dec_seq_len = dec_step
                break
        assert dec_seq_len > 0
        print("All outputs seem to match.")
    else:
        print(
            "blocks_debug_dump_output not specified. It will not compare the model outputs." % blocks_debug_dump_output
        )

    if dry_run:
        print("Dry-run, not saving model.")
    else:
        rnn.engine.save_model(our_model_fn)
    print("Finished importing.")


def lstm_vec_blocks_to_tf(x):
    """
    Blocks order: gate-in, gate-forget, next-in, gate-out
    TF order: i = input_gate, j = new_input, f = forget_gate, o = output_gate
    :param numpy.ndarray x: (..., dim*4)
    :rtype: numpy.ndarray
    """
    axis = x.ndim - 1
    i, f, j, o = numpy.split(x, 4, axis=axis)
    return numpy.concatenate([i, j, f, o], axis=axis)


def calc_lstm(x, blocks_params, t=0):
    """
    :param numpy.ndarray x: (seq_len, in_dim)
    :param dict[str,numpy.ndarray] blocks_params:
    :param int t:
    :rtype: numpy.ndarray
    """
    prefix = "bidirectionalencoder/EncoderBidirectionalLSTM1"
    prefix1 = "%s/bidirectionalseparateparameters/forward." % prefix
    prefix2 = "%s/fwd_fork/fork_inputs." % prefix
    W_in, b = blocks_params[prefix2 + "W"], blocks_params[prefix2 + "b"]
    assert b.ndim == 1
    assert b.shape[0] % 4 == 0
    out_dim = b.shape[0] // 4
    seq_len, in_dim = x.shape
    assert W_in.shape == (in_dim, out_dim * 4)
    z = numpy.dot(x[t], W_in) + b
    assert z.shape == (out_dim * 4,)
    cur_state, cur_cell = calc_raw_lstm(z, blocks_params=blocks_params, prefix=prefix1)
    return cur_state


def calc_raw_lstm(z, blocks_params, prefix, last_state=None, last_cell=None):
    """
    :param numpy.ndarray z: shape (out_dim * 4,)
    :param dict[str,numpy.ndarray] blocks_params:
    :param str prefix: e.g. "bidirectionalencoder/EncoderBidirectionalLSTM1/bidirectionalseparateparameters/forward."
    :param numpy.ndarray|None last_state: (out_dim,)
    :param numpy.ndarray|None last_cell: (out_dim,)
    :rtype: numpy.ndarray
    """
    assert z.ndim == 1
    assert z.shape[-1] % 4 == 0
    out_dim = z.shape[-1] // 4
    W_re = blocks_params[prefix + "W_state"]
    assert W_re.shape == (out_dim, out_dim * 4)
    W_cell_to_in = blocks_params[prefix + "W_cell_to_in"]
    W_cell_to_forget = blocks_params[prefix + "W_cell_to_forget"]
    W_cell_to_out = blocks_params[prefix + "W_cell_to_out"]
    assert W_cell_to_in.shape == W_cell_to_forget.shape == W_cell_to_out.shape == (out_dim,)
    if last_cell is None and last_state is None:
        initial_state, initial_cell = blocks_params[prefix + "initial_state"], blocks_params[prefix + "initial_cells"]
        assert initial_state.shape == initial_cell.shape == (out_dim,)
        last_state = initial_state
        last_cell = initial_cell
    z = z + numpy.dot(last_state, W_re)
    assert z.shape == (out_dim * 4,)
    # Blocks order: gate-in, gate-forget, next-in, gate-out
    gate_in, gate_forget, next_in, gate_out = numpy.split(z, 4)
    gate_in = gate_in + W_cell_to_in * last_cell
    gate_forget = gate_forget + W_cell_to_forget * last_cell
    gate_in = sigmoid(gate_in)
    gate_forget = sigmoid(gate_forget)
    next_in = numpy.tanh(next_in)
    cur_cell = last_cell * gate_forget + next_in * gate_in
    gate_out = gate_out + W_cell_to_out * cur_cell
    gate_out = sigmoid(gate_out)
    cur_state = numpy.tanh(cur_cell) * gate_out
    return cur_state, cur_cell


def softmax(x, axis=-1):
    assert isinstance(x, numpy.ndarray)
    e_x = numpy.exp(x - numpy.max(x, axis=axis, keepdims=True))
    assert isinstance(e_x, numpy.ndarray)
    return e_x / e_x.sum(axis=axis, keepdims=True)


def log_softmax(x, axis=-1):
    assert isinstance(x, numpy.ndarray)
    xdev = x - x.max(axis=axis, keepdims=True)
    lsm = xdev - numpy.log(numpy.sum(numpy.exp(xdev), axis=axis, keepdims=True))
    return lsm


def sigmoid(x):
    return 1.0 / (1.0 + numpy.exp(-x))


if __name__ == "__main__":
    better_exchook.install()
    main()



"""
tests for returnn.tensor
"""

import _setup_test_env  # noqa
import sys
import unittest
from returnn.util import better_exchook
from returnn.tensor import Tensor, Dim


def test_tensor():
    batch_dim = Dim(name="batch", dimension=None)
    time_dim = Dim(name="time", dimension=None)
    feat_dim = Dim(10)
    x = Tensor("x", (batch_dim, time_dim, feat_dim), "float32")
    print(x)


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Import some of the torch.nn modules.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import returnn.frontend as rf

if TYPE_CHECKING:
    import torch


def import_params_torch_linear_to_rf(model_pt: torch.nn.Linear, model_rf: rf.Linear):
    """
    import params from torch.nn.Linear to rf.Linear
    """
    import torch

    assert isinstance(model_pt, torch.nn.Linear)
    assert isinstance(model_rf, rf.Linear)
    assert model_rf.with_bias == (model_pt.bias is not None)

    with torch.no_grad():
        model_rf.weight.raw_tensor.copy_(model_pt.weight.T)  # (in,out)
        if model_rf.with_bias:
            model_rf.bias.raw_tensor.copy_(model_pt.bias)  # (out,)


def import_params_torch_conv1d_to_rf(model_pt: torch.nn.Conv1d, model_rf: rf.Conv1d):
    """
    import params from torch.nn.Conv1d to rf.Conv1d
    """
    import torch

    assert isinstance(model_pt, torch.nn.Conv1d)
    assert isinstance(model_rf, rf.Conv1d)
    assert model_rf.with_bias == (model_pt.bias is not None)

    with torch.no_grad():
        # Torch shape: out_channels, in_channels // groups, *kernel_size
        # RF shape: self.out_dim, self.filter_in_dim, *self.filter_size, i.e. should be same
        model_rf.filter.raw_tensor.copy_(model_pt.weight)
        if model_rf.with_bias:
            model_rf.bias.raw_tensor.copy_(model_pt.bias)


def import_params_torch_layer_norm_to_rf(model_pt: torch.nn.LayerNorm, model_rf: rf.LayerNorm):
    """
    Import the parameters from torch.nn.LayerNorm to rf.LayerNorm.
    """
    import torch

    assert isinstance(model_pt, torch.nn.LayerNorm)
    assert isinstance(model_rf, rf.LayerNorm)
    assert model_pt.weight.shape[0] == model_rf.in_dim.dimension

    with torch.no_grad():
        model_rf.scale.raw_tensor.copy_(model_pt.weight)  # (in,)
        model_rf.bias.raw_tensor.copy_(model_pt.bias)  # (in,)

    num_params_pt = 0
    for k, v in model_pt.named_parameters():
        num_params_pt += v.numel()
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    assert num_params_rf == num_params_pt



"""
Some helper utils.
"""

from __future__ import annotations
from typing import Optional, Union, Sequence, Dict, List, Tuple
import numpy
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim


def tensor_dict_fill_random_numpy_(
    tensor_dict: TensorDict,
    *,
    rnd: Union[int, numpy.random.RandomState] = 42,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
):
    """
    Random fill with NumPy arrays.

    :param tensor_dict:
    :param rnd:
    :param dyn_dim_max_sizes: you can specify max sizes for dim tags with dynamic sizes.
        The fill random code makes sure that there is at least one entry where we reach the max size,
        so that the dim value will be the max size.
    :param dyn_dim_min_sizes:
    """
    if not isinstance(rnd, numpy.random.RandomState):
        rnd = numpy.random.RandomState(rnd)
    for v in tensor_dict.data.values():
        tensor_fill_random_numpy_(v, rnd=rnd, dyn_dim_max_sizes=dyn_dim_max_sizes, dyn_dim_min_sizes=dyn_dim_min_sizes)


def tensor_fill_random_numpy_(
    x: Tensor,
    *,
    min_val: int = 0,
    max_val: Optional[int] = None,
    rnd: numpy.random.RandomState,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
) -> bool:
    """fill. return whether sth was filled"""
    if dyn_dim_max_sizes is None:
        dyn_dim_max_sizes = {}
    if dyn_dim_min_sizes is None:
        dyn_dim_min_sizes = {}
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if dim.is_batch_dim() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor("batch", [], dtype="int32")
            if dim.is_dynamic() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor(dim.name or "time", dims=[batch_dim], dtype="int32")
            if not dim.dyn_size_ext:
                continue
            if tensor_fill_random_numpy_(
                dim.dyn_size_ext,
                min_val=dyn_dim_min_sizes.get(dim, 2),
                max_val=dyn_dim_max_sizes.get(dim, None),
                rnd=rnd,
                dyn_dim_max_sizes=dyn_dim_max_sizes,
            ):
                if dim in dyn_dim_max_sizes:
                    # Make sure at least one of the dyn sizes matches the max size.
                    i = rnd.randint(0, dim.dyn_size_ext.raw_tensor.size)
                    dim.dyn_size_ext.raw_tensor.flat[i] = dyn_dim_max_sizes[dim]
                    if dim in dyn_dim_min_sizes:
                        j = rnd.randint(0, dim.dyn_size_ext.raw_tensor.size - 1)
                        if j >= i:
                            j += 1
                        dim.dyn_size_ext.raw_tensor.flat[j] = dyn_dim_min_sizes[dim]
                elif dim in dyn_dim_min_sizes:
                    raise Exception(f"also define {dim} in dyn_dim_max_sizes, not just dyn_dim_min_sizes")
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True
            elif not isinstance(dim.dyn_size_ext.raw_tensor, numpy.ndarray):
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, numpy.ndarray):
            x.raw_tensor = None

    if x.raw_tensor is None:
        shape = [d.get_dim_value() for d in x.dims]
        if x.dtype.startswith("int"):
            if max_val is None:
                max_val = rnd.randint(5, 20)
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            x.raw_tensor = rnd.randint(min_val, max_val, size=shape, dtype=x.dtype)
        elif x.dtype == "bool":
            x.raw_tensor = rnd.randint(0, 2, size=shape, dtype=x.dtype)
        elif x.dtype.startswith("float"):
            x.raw_tensor = rnd.normal(0.0, 1.0, size=shape).astype(x.dtype)
        elif x.dtype == "bfloat16":
            # Numpy does not support bfloat16, will later be casted to bfloat16
            x.raw_tensor = rnd.normal(0.0, 1.0, size=shape).astype("float32")
        elif x.dtype.startswith("complex"):
            real = rnd.normal(0.0, 1.0, size=shape)
            imag = rnd.normal(0.0, 1.0, size=shape)
            x.raw_tensor = (real + 1j * imag).astype(x.dtype)
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        filled = True

    assert isinstance(x.raw_tensor, numpy.ndarray)

    return filled


def tensor_dict_dims_random_seq_len_min_max(
    tensor_dict: TensorDict,
    seq_lens: Union[None, int, Tuple[int, int], Dict[Union[str, Dim, None], Union[int, Tuple[int, int]]]] = None,
) -> Tuple[List[Dim], Dict[Dim, Tuple[int, int]]]:
    """
    This is specifically intended to prepare the list of all dynamic dims from the tensor dict
    and the seq_len_min_max for :func:`get_random_seq_lens_for_dyn_dims`.

    :param tensor_dict:
    :param seq_lens: either fixed seq len, or take randint. per data key, or per dim, or same for all
    :return: dims, seq_len_min_max
    """
    if seq_lens is None:
        seq_lens = {}
    if not isinstance(seq_lens, dict):
        seq_lens = {None: seq_lens}
    seq_lens: Dict[Union[str, Dim, None], Union[int, Tuple[int, int]]]

    # Collect all dyn dim tags, including derived_from_op ones.
    # The order will be sorted such that derived_from_op roots come first.
    visited_dims = set()
    dims = []
    seq_len_min_max = {}  # Also collect seq_len_min_max.
    for k, v in tensor_dict.data.items():
        for dim in v.dims:
            if dim.is_dynamic() and dim not in visited_dims and not dim.is_batch_dim():
                queue = [dim]
                offset = len(dims)
                while queue:
                    dim = queue.pop(0)
                    if not dim.is_dynamic():
                        continue
                    if dim in visited_dims:
                        continue
                    visited_dims.add(dim)
                    dims.insert(offset, dim)
                    dim.reset_batch_and_raw()
                    if dim.derived_from_op:
                        queue.extend(dim.derived_from_op.inputs)
                    else:
                        # Need to specify seq_len_min_max.
                        if dim in seq_lens or k in seq_lens or None in seq_lens:
                            if dim in seq_lens:
                                size = seq_lens[dim]
                            elif k in seq_lens:
                                size = seq_lens[k]
                            else:
                                size = seq_lens[None]
                            if isinstance(size, int):
                                size = (size, size)
                            else:
                                assert (
                                    isinstance(size, tuple)
                                    and len(size) == 2
                                    and all(isinstance(s, int) for s in size)
                                    and 0 <= size[0] <= size[1]
                                ), f"invalid size {size!r} in seq lens {seq_lens}"
                        else:
                            if v.shape in {(None,), (None, 1)} and v.dtype.startswith("float"):
                                # Assume raw audio data samples, take longer seq lens by default, assume 16khz.
                                size = (1_000, 8_000)
                            else:
                                size = (5, 15)
                        seq_len_min_max[dim] = size

    return dims, seq_len_min_max


def get_random_seq_lens_for_dyn_dims(
    dims: Sequence[Dim],
    seq_len_min_max: Dict[Dim, Tuple[int, int]],
    *,
    batch_size: int = 1,
    rnd: Union[int, numpy.random.RandomState] = 1337,
) -> Dict[Dim, numpy.ndarray]:
    """
    Make random seq lens for dims.

    Note that dim tags are not actually modified here,
    as we need to have this in a safe way,
    which might run in parallel to the main thread.

    :param dims: Note that the order matter, as we use complete_dyn_size() (or equivalent).
    :param seq_len_min_max:
    :param batch_size:
    :param rnd:
    """
    if not isinstance(rnd, numpy.random.RandomState):
        rnd = numpy.random.RandomState(rnd)

    gen_dims = {}
    for dim in dims:
        if dim not in gen_dims:
            if dim.derived_from_op:
                # If we get a KeyError for the following, the order of dims is invalid.
                values = [gen_dims[dim_] for dim_ in dim.derived_from_op.inputs]
                kind = dim.derived_from_op.kind
                a = values[0]
                for b in values[1:]:
                    if kind == "add":
                        a = numpy.maximum(a + b, 0)
                    elif kind == "sub":
                        a = numpy.maximum(a - b, 0)
                    elif kind == "mul":
                        a = a * b
                    elif kind in ("floordiv", "truediv"):  # truediv assumes there is no remainder
                        a = a // b
                    elif kind == "ceildiv":
                        a = -(-a // b)
                    else:
                        raise ValueError("unknown op kind %r" % kind)
                gen_dims[dim] = a
                continue

            min_, max_ = seq_len_min_max[dim]
            gen_dims[dim] = rnd.randint(min_, max_ + 1, size=[batch_size], dtype=numpy.int32)

    return gen_dims



#!/usr/bin/env python3
# coding=utf-8
# Copyright 2020 The Tensor2Tensor Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to average values of variables in a list of checkpoint files.

This script is generic for any TF checkpoint. It is not specific to RETURNN.

Original code:
https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/utils/avg_checkpoints.py
"""

from __future__ import annotations

import os
import numpy
import logging
import tensorflow as tf

import _setup_returnn_env  # noqa
import returnn.tf.compat as tf_compat
from returnn.util import better_exchook

better_exchook.install()

flags = tf_compat.v1.flags
FLAGS = flags.FLAGS

flags.DEFINE_string("checkpoints", "", "Comma-separated list of checkpoints to average.")
flags.DEFINE_integer(
    "num_last_checkpoints", 0, "Averages the last N saved checkpoints. If the checkpoints flag is set, this is ignored."
)
flags.DEFINE_string("prefix", "", "Prefix (e.g., directory) to append to each checkpoint.")
flags.DEFINE_string("output_path", "/tmp/averaged.ckpt", "Path to output the averaged checkpoint to.")


def checkpoint_exists(path):
    """
    :param str path:
    :rtype: bool
    """
    return (
        tf_compat.v1.gfile.Exists(path)
        or tf_compat.v1.gfile.Exists(path + ".meta")
        or tf_compat.v1.gfile.Exists(path + ".index")
    )


def main(_):
    """
    Main entry.
    """
    _logger = logging.getLogger("tensorflow")
    _logger.setLevel("INFO")
    tf_compat.v1.logging.info("%s startup. TF version: %s" % (__file__, tf.__version__))

    if FLAGS.checkpoints:
        # Get the checkpoints list from flags and run some basic checks.
        checkpoints = [c.strip() for c in FLAGS.checkpoints.split(",")]
        checkpoints = [c for c in checkpoints if c]
        if not checkpoints:
            raise ValueError("No checkpoints provided for averaging.")
        if FLAGS.prefix:
            checkpoints = [FLAGS.prefix + c for c in checkpoints]
    else:
        assert FLAGS.num_last_checkpoints >= 1, "Must average at least one model"
        assert FLAGS.prefix, "Prefix must be provided when averaging last N checkpoints"
        checkpoint_state = tf.train.get_checkpoint_state(os.path.dirname(FLAGS.prefix))
        # Checkpoints are ordered from oldest to newest.
        checkpoints = checkpoint_state.all_model_checkpoint_paths[-FLAGS.num_last_checkpoints :]

    checkpoints = [c for c in checkpoints if checkpoint_exists(c)]
    if not checkpoints:
        if FLAGS.checkpoints:
            raise ValueError("None of the provided checkpoints exist. %s" % FLAGS.checkpoints)
        else:
            raise ValueError("Could not find checkpoints at %s" % os.path.dirname(FLAGS.prefix))

    # Read variables from all checkpoints and average them.
    tf_compat.v1.logging.info("Reading variables and averaging checkpoints:")
    for c in checkpoints:
        tf_compat.v1.logging.info("%s ", c)
    var_list = tf.train.list_variables(checkpoints[0])
    var_values, var_dtypes = {}, {}
    for (name, shape) in var_list:
        var_values[name] = numpy.zeros(shape)
    for checkpoint in checkpoints:
        reader = tf.train.load_checkpoint(checkpoint)
        for name in var_values:
            tensor = reader.get_tensor(name)
            if not isinstance(tensor, numpy.ndarray):  # e.g. int (scalar)
                tensor = numpy.array(tensor)
            assert isinstance(tensor, numpy.ndarray)
            var_dtypes[name] = tensor.dtype
            if isinstance(tensor.dtype, numpy.integer):
                var_values[name] = tensor  # just take last
            else:
                var_values[name] += tensor
        tf_compat.v1.logging.info("Read from checkpoint %s", checkpoint)
    for name in var_values:  # Average.
        if not isinstance(var_dtypes[name], numpy.integer):
            var_values[name] /= len(checkpoints)

    with tf_compat.v1.variable_scope(tf_compat.v1.get_variable_scope(), reuse=tf_compat.v1.AUTO_REUSE):
        tf_vars = [tf_compat.v1.get_variable(v, shape=var_values[v].shape, dtype=var_dtypes[v]) for v in var_values]
    placeholders = [tf_compat.v1.placeholder(v.dtype, shape=v.shape) for v in tf_vars]
    assign_ops = [tf_compat.v1.assign(v, p) for (v, p) in zip(tf_vars, placeholders)]
    saver = tf_compat.v1.train.Saver(tf_compat.v1.all_variables())

    # Build a model consisting only of variables, set them to the average values.
    with tf_compat.v1.Session() as sess:
        sess.run(tf_compat.v1.global_variables_initializer())
        for p, assign_op, (name, value) in zip(placeholders, assign_ops, var_values.items()):
            sess.run(assign_op, {p: value})
        # Use the built saver to save the averaged checkpoint.
        saver.save(sess, FLAGS.output_path)

    tf_compat.v1.logging.info("Averaged checkpoints saved in %s", FLAGS.output_path)


if __name__ == "__main__":
    tf_compat.v1.app.run()
